# coding=utf-8
#
# Copyright (C) 2018 Martin Owens <doctormo@gmail.com>
# Copyright (C) 2023 Jonathan Neuhauser <jonathan.neuhauser@outlook.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""Line-like path commands (Line, Horz, Vert, ZoneClose and their relative siblings)"""

from __future__ import annotations

from typing import overload, Tuple, Optional, TYPE_CHECKING, Callable, Union

from inkex.paths.interfaces import ILengthSettings, LengthSettings

from ..transforms import Transform, BoundingBox, ComplexLike

from .interfaces import AbsolutePathCommand, RelativePathCommand, ILengthSettings

if TYPE_CHECKING:
    from .curves import Curve


class LineMixin:
    """Common Line functions"""

    # pylint: disable=unused-argument

    arg1: complex

    cend_point: Callable[[complex, complex], complex]

    def ccurve_points(self, first: complex, prev: complex, prev_prev: complex):
        """Common implementation of ccurve_points for Lines"""
        arg1 = self.cend_point(first, prev)
        return prev, arg1, arg1

    def ccontrol_points(
        self,
        first: complex,
        prev: complex,
        prev_prev: complex,  # pylint: disable=unused-argument
    ) -> Tuple[complex, ...]:
        """Common implementation of ccontrol_points for Lines"""
        return (self.cend_point(first, prev),)

    def _cderivative(
        self, first: complex, prev: complex, prev_control: complex, t: float, n: int = 1
    ) -> complex:
        start = self.cend_point(first, prev)
        if prev == start:
            raise ValueError("Derivative is not defined for zero-length segments")
        if n == 1:
            return start - prev
        return 0j

    def _curvature(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ) -> float:
        return 0

    def _cpoint(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ) -> complex:
        return self.cend_point(first, prev) * t + (1 - t) * prev

    def _length(
        self,
        first: complex,
        prev: complex,
        prev_control: complex,
        t0: float = 0,
        t1: float = 1,
        settings=LengthSettings(),
    ) -> float:
        return abs(self.cend_point(first, prev) - prev) * (t1 - t0)

    def _ilength(
        self,
        first: complex,
        prev: complex,
        prev_control: complex,
        length: float,
        settings: ILengthSettings = ILengthSettings(),
    ):
        return length / self._length(first, prev, prev_control)


class Line(LineMixin, AbsolutePathCommand):
    """Line segment"""

    letter = "L"
    nargs = 2

    arg1: complex
    """The (absolute) end points of the line."""

    @property
    def x(self):
        """x coordinate of the line's (absolute) end point."""
        return self.arg1.real

    @property
    def y(self):
        """y coordinate of the line's (absolute) end point."""
        return self.arg1.imag

    @property
    def args(self):
        return self.x, self.y

    @overload
    def __init__(self, x: ComplexLike): ...

    @overload
    def __init__(self, x: float, y: float): ...

    def __init__(self, x, y=None):
        if y is not None:
            self.arg1 = x + y * 1j
        else:
            self.arg1 = complex(x)

    def update_bounding_box(self, first, last_two_points, bbox):
        bbox += BoundingBox(
            (last_two_points[-1].real, self.x), (last_two_points[-1].imag, self.y)
        )

    def to_relative(self, prev: ComplexLike) -> line:
        return line(self.arg1 - prev)

    def transform(self, transform) -> Line:
        return Line(transform.capply_to_point(self.arg1))

    def cend_point(self, first: complex, prev: complex) -> complex:
        # pylint: disable=unused-argument
        return self.arg1

    def reverse(self, first: ComplexLike, prev: ComplexLike) -> Line:
        return Line(prev)

    def _split(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ) -> Tuple[Line, Line]:
        return Line(self._cpoint(first, prev, prev_control, t)), Line(self.arg1)


class line(LineMixin, RelativePathCommand):  # pylint: disable=invalid-name
    """Relative line segment"""

    letter = "l"
    nargs = 2

    arg1: complex
    """The (relative) end points of the line."""

    @property
    def dx(self):
        """x coordinate of the line's (relative) end point."""
        return self.arg1.real

    @property
    def dy(self):
        """y coordinate of the line's (relative) end point."""
        return self.arg1.imag

    @property
    def args(self):
        return self.dx, self.dy

    @overload
    def __init__(self, dx: ComplexLike): ...

    @overload
    def __init__(self, dx: float, dy: float): ...

    def __init__(self, dx, dy=None):
        if dy is not None:
            self.arg1 = dx + dy * 1j
        else:
            self.arg1 = complex(dx)

    def to_absolute(self, prev: ComplexLike) -> Line:
        return Line(prev + self.arg1)

    def cend_point(self, first: complex, prev: complex) -> complex:
        # pylint: disable=unused-argument
        return self.arg1 + prev

    def reverse(self, first: ComplexLike, prev: ComplexLike) -> line:
        return line(-self.arg1)

    def to_curve(
        self, prev: ComplexLike, prev_prev: Optional[ComplexLike] = 0j
    ) -> Curve:
        raise ValueError("Move segments can not be changed into curves.")

    def _split(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ) -> Tuple[line, line]:
        dx1 = self.cpoint(first, prev, prev_control, t) - prev
        return line(dx1), line(self.arg1 - dx1)


class MoveMixin:
    """Disable derivative / length method for Move command."""

    def _cderivative(
        self, first: complex, prev: complex, prev_control: complex, t: float, n: int = 1
    ) -> complex:
        raise ValueError("Derivative is not supported for move/Move")

    def _cunit_tangent(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ) -> complex:
        raise ValueError("Unit Tangent is not supported for move/Move")

    def _curvature(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ) -> float:
        raise ValueError("Curvature is not supported for move/Move")

    def _cpoint(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ) -> complex:
        raise ValueError("Point is not supported for move/Move")

    def _length(
        self,
        first: complex,
        prev: complex,
        prev_control: complex,
        t0: float = 0,
        t1: float = 1,
        settings=LengthSettings(),
    ) -> float:
        raise ValueError("Length is not supported for move/Move")

    def _ilength(
        self,
        first: complex,
        prev: complex,
        prev_control: complex,
        length: float,
        settings: ILengthSettings = ILengthSettings(),
    ):
        raise ValueError("ILength is not supported for move/Move")

    def _split(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ) -> Tuple[Move, Move]:
        raise ValueError("Split is not supported for move/Move")


class Move(MoveMixin, AbsolutePathCommand):
    """Move pen segment without a line"""

    letter = "M"
    nargs = 2
    next_command = Line

    arg1: complex
    """The (absolute) end points of the Move command"""

    @property
    def x(self):
        """x coordinate of the Moves's (absolute) end point."""
        return self.arg1.real

    @property
    def y(self):
        """y coordinate of the Move's (absolute) end point."""
        return self.arg1.imag

    @property
    def args(self):
        return self.x, self.y

    @overload
    def __init__(self, x: ComplexLike): ...

    @overload
    def __init__(self, x: float, y: float): ...

    def __init__(self, x, y=None):
        if y is not None:
            self.arg1 = x + y * 1j
        else:
            self.arg1 = complex(x)

    def update_bounding_box(self, first, last_two_points, bbox):
        bbox += BoundingBox(self.x, self.y)

    def ccurve_points(self, first: complex, prev: complex, prev_prev: complex):
        return prev, self.arg1, self.arg1

    def ccontrol_points(
        self, first: complex, prev: complex, prev_prev: complex
    ) -> Tuple[complex, ...]:
        return (self.arg1,)

    def to_relative(self, prev: ComplexLike) -> move:
        return move(self.arg1 - prev)

    def transform(self, transform: Transform) -> Move:
        return Move(transform.capply_to_point(self.arg1))

    def cend_point(self, first: complex, prev: complex) -> complex:
        return self.arg1

    def to_curve(
        self, prev: ComplexLike, prev_prev: Optional[ComplexLike] = 0j
    ) -> Curve:
        raise ValueError("Move segments can not be changed into curves.")

    def reverse(self, first: ComplexLike, prev: ComplexLike) -> Move:
        return Move(prev)


class move(MoveMixin, RelativePathCommand):  # pylint: disable=invalid-name
    """Relative move segment"""

    letter = "m"
    nargs = 2
    next_command = line

    @property
    def dx(self):
        """x coordinate of the moves's (relative) end point."""
        return self.arg1.real

    @property
    def dy(self):
        """y coordinate of the move's (relative) end point."""
        return self.arg1.imag

    @property
    def args(self):
        return self.dx, self.dy

    @overload
    def __init__(self, dx: ComplexLike): ...

    @overload
    def __init__(self, dx: float, dy: float): ...

    def __init__(self, dx, dy=None):
        if dy is not None:
            self.arg1 = dx + dy * 1j
        else:
            self.arg1 = complex(dx)

    def ccurve_points(self, first: complex, prev: complex, prev_prev: complex):
        return prev, self.arg1 + prev, self.arg1 + prev

    def ccontrol_points(
        self, first: complex, prev: complex, prev_prev: complex
    ) -> Tuple[complex, ...]:
        return (self.arg1 + prev,)

    def cend_point(self, first: complex, prev: complex) -> complex:
        return self.arg1 + prev

    def to_absolute(self, prev: ComplexLike) -> Move:
        return Move(prev + self.arg1)

    def reverse(self, first: ComplexLike, prev: ComplexLike) -> move:
        return move(prev - first)

    def to_curve(
        self, prev: ComplexLike, prev_prev: Optional[ComplexLike] = 0j
    ) -> Curve:
        raise ValueError("Move segments can not be changed into curves.")


class ZoneClose(LineMixin, AbsolutePathCommand):
    """Close segment to finish a path"""

    letter = "Z"
    nargs = 0
    next_command = Move

    @property
    def args(self):
        return ()

    def update_bounding_box(self, first, last_two_points, bbox):
        pass

    def transform(self, transform: Transform) -> ZoneClose:
        return ZoneClose()

    def to_relative(self, prev: ComplexLike) -> zoneClose:
        return zoneClose()

    def cend_point(self, first: complex, prev: complex) -> complex:
        # pylint: disable=unused-argument
        return first

    def to_curve(
        self, prev: ComplexLike, prev_prev: Optional[ComplexLike] = 0j
    ) -> Curve:
        raise ValueError("ZoneClose segments can not be changed into curves.")

    def reverse(self, first: ComplexLike, prev: ComplexLike) -> Line:
        return Line(prev)

    def _split(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ) -> Tuple[Line, ZoneClose]:
        return Line(self._cpoint(first, prev, prev_control, t)), ZoneClose()


class zoneClose(LineMixin, RelativePathCommand):  # pylint: disable=invalid-name
    """Same as above (svg says no difference)"""

    letter = "z"
    nargs = 0
    next_command = Move

    @property
    def args(self):
        return ()

    def to_absolute(self, prev: ComplexLike):
        return ZoneClose()

    def reverse(self, first: ComplexLike, prev: ComplexLike) -> line:
        return line(prev - first)

    def cend_point(self, first: complex, prev: complex) -> complex:
        # pylint: disable=unused-argument
        return first

    def to_curve(
        self, prev: ComplexLike, prev_prev: Optional[ComplexLike] = 0j
    ) -> Curve:
        raise ValueError("ZoneClose segments can not be changed into curves.")

    def _split(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ) -> Tuple[line, zoneClose]:
        return line(self.cpoint(first, prev, prev_control, t) - prev), zoneClose()


class Horz(LineMixin, AbsolutePathCommand):
    """Horizontal Line segment"""

    letter = "H"
    nargs = 1

    @property
    def args(self):
        return (self.x,)

    def __init__(self, x):
        self.x = x

    def update_bounding_box(self, first, last_two_points, bbox):
        bbox += BoundingBox(
            (last_two_points[-1].real, self.x), last_two_points[-1].imag
        )

    def to_relative(self, prev: ComplexLike) -> horz:
        return horz(self.x - complex(prev).real)

    def to_non_shorthand(self, prev: ComplexLike, prev_control: ComplexLike) -> Line:
        return self.to_line(prev)

    def transform(self, transform: Transform) -> AbsolutePathCommand:
        raise ValueError("Horizontal lines can't be transformed directly.")

    def cend_point(self, first: complex, prev: complex) -> complex:
        # pylint: disable=unused-argument
        return self.x + prev.imag * 1j

    def reverse(self, first: ComplexLike, prev: ComplexLike) -> Horz:
        return Horz(complex(prev).real)

    def _split(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ) -> Tuple[Horz, Horz]:
        return Horz(self.cpoint(first, prev, prev_control, t).real), Horz(self.x)


class horz(LineMixin, RelativePathCommand):  # pylint: disable=invalid-name
    """Relative horizontal line segment"""

    letter = "h"
    nargs = 1

    @property
    def args(self):
        return (self.dx,)

    def __init__(self, dx):
        self.dx = dx

    def to_absolute(self, prev: ComplexLike) -> Horz:
        return Horz(complex(prev).real + self.dx)

    def to_non_shorthand(self, prev: ComplexLike, prev_control: ComplexLike) -> Line:
        return self.to_line(prev)

    def cend_point(self, first: complex, prev: complex) -> complex:
        # pylint: disable=unused-argument
        return (self.dx + prev.real) + prev.imag * 1j

    def reverse(self, first: ComplexLike, prev: ComplexLike) -> horz:
        return horz(-self.dx)

    def _split(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ) -> Tuple[horz, horz]:
        dx1 = (self.cpoint(first, prev, prev_control, t) - prev).real
        return horz(dx1), horz(self.dx - dx1)


class Vert(LineMixin, AbsolutePathCommand):
    """Vertical Line segment"""

    letter = "V"
    nargs = 1

    @property
    def args(self):
        return (self.y,)

    def __init__(self, y):
        self.y = y

    def update_bounding_box(self, first, last_two_points, bbox):
        bbox += BoundingBox(
            last_two_points[-1].real, (last_two_points[-1].imag, self.y)
        )

    def transform(self, transform: Transform) -> AbsolutePathCommand:
        raise ValueError("Vertical lines can't be transformed directly.")

    def to_non_shorthand(self, prev: ComplexLike, prev_control: ComplexLike) -> Line:
        return self.to_line(prev)

    def to_relative(self, prev: ComplexLike) -> vert:
        return vert(self.y - complex(prev).imag)

    def cend_point(self, first: complex, prev: complex) -> complex:
        # pylint: disable=unused-argument
        return prev.real + self.y * 1j

    def reverse(self, first: ComplexLike, prev: ComplexLike) -> Vert:
        return Vert(complex(prev).imag)

    def _split(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ) -> Tuple[Vert, Vert]:
        return Vert(self.cpoint(first, prev, prev_control, t).imag), Vert(self.y)


class vert(LineMixin, RelativePathCommand):  # pylint: disable=invalid-name
    """Relative vertical line segment"""

    letter = "v"
    nargs = 1

    @property
    def args(self):
        return (self.dy,)

    def __init__(self, dy):
        self.dy = dy

    def to_absolute(self, prev: ComplexLike) -> Vert:
        return Vert(complex(prev).imag + self.dy)

    def to_non_shorthand(self, prev: ComplexLike, prev_control: ComplexLike) -> Line:
        return self.to_line(prev)

    def cend_point(self, first: complex, prev: complex) -> complex:
        # pylint: disable=unused-argument
        return prev.real + (prev.imag + self.dy) * 1j

    def reverse(self, first: ComplexLike, prev: ComplexLike) -> vert:
        return vert(-self.dy)

    def _split(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ) -> Tuple[vert, vert]:
        dy1 = (self.cpoint(first, prev, prev_control, t) - prev).imag
        return vert(dy1), vert(self.dy - dy1)



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2014-2019 Martin Owens
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
Merges styles into class based styles and removes.
"""

import inkex
from inkex.localization import inkex_gettext as _


class MergeStyles(inkex.EffectExtension):
    """Merge any styles which are the same for CSS"""

    def add_arguments(self, pars):
        self.arg_parser.add_argument(
            "-n",
            "--name",
            type=str,
            dest="name",
            help="Name of selected element's common class",
        )

    def effect(self):
        """Apply the style effect"""
        newclass = self.options.name
        if not newclass:
            newclass = self.svg.get_unique_id("css")

        elements = self.svg.selection.values()
        common = None

        for elem in elements:
            style = set(elem.style.items())
            if common is not None:
                common &= style
            else:
                common = style

        if not common:
            return inkex.errormsg(
                _("There are no common styles between these elements.")
            )
        self.svg.stylesheet.add("." + newclass, inkex.Style(sorted(common)))

        for elem in elements:
            elem.style -= dict(common).keys()
            elem.classes.append(newclass)
        return True


if __name__ == "__main__":
    MergeStyles().run()



# Forked from cssselect2, 1.2.1, BSD License

"""Parse CSS declarations."""

from tinycss2 import parse_component_value_list

__all__ = ["parse"]

SUPPORTED_PSEUDO_ELEMENTS = {
    # As per CSS Pseudo-Elements Module Level 4
    "first-line",
    "first-letter",
    "prefix",
    "postfix",
    "selection",
    "target-text",
    "spelling-error",
    "grammar-error",
    "before",
    "after",
    "marker",
    "placeholder",
    "file-selector-button",
    # As per CSS Generated Content for Paged Media Module
    "footnote-call",
    "footnote-marker",
    # As per CSS Scoping Module Level 1
    "content",
    "shadow",
}


def parse(input, namespaces=None, forgiving=False, relative=False):
    """Yield tinycss2 selectors found in given ``input``.

    :param input:
        A string, or an iterable of tinycss2 component values.

    """
    if isinstance(input, str):
        input = parse_component_value_list(input)
    tokens = TokenStream(input)
    namespaces = namespaces or {}
    try:
        yield parse_selector(tokens, namespaces, relative)
    except SelectorError as exception:
        if forgiving:
            return
        raise exception
    while 1:
        next = tokens.next()
        if next is None:
            return
        elif next == ",":
            try:
                yield parse_selector(tokens, namespaces, relative)
            except SelectorError as exception:
                if not forgiving:
                    raise exception
        else:
            if not forgiving:
                raise SelectorError(next, f"unexpected {next.type} token.")


def parse_selector(tokens, namespaces, relative=False):
    tokens.skip_whitespace_and_comment()
    if relative:
        peek = tokens.peek()
        if peek in (">", "+", "~"):
            initial_combinator = peek.value
            tokens.next()
        else:
            initial_combinator = " "
        tokens.skip_whitespace_and_comment()
    result, pseudo_element = parse_compound_selector(tokens, namespaces)
    while 1:
        has_whitespace = tokens.skip_whitespace()
        while tokens.skip_comment():
            has_whitespace = tokens.skip_whitespace() or has_whitespace
        selector = Selector(result, pseudo_element)
        if relative:
            selector = RelativeSelector(initial_combinator, selector)
        if pseudo_element is not None:
            return selector
        peek = tokens.peek()
        if peek is None or peek == ",":
            return selector
        elif peek in (">", "+", "~"):
            combinator = peek.value
            tokens.next()
        elif has_whitespace:
            combinator = " "
        else:
            return selector
        compound, pseudo_element = parse_compound_selector(tokens, namespaces)
        result = CombinedSelector(result, combinator, compound)


def parse_compound_selector(tokens, namespaces):
    type_selectors = parse_type_selector(tokens, namespaces)
    simple_selectors = type_selectors if type_selectors is not None else []
    while 1:
        simple_selector, pseudo_element = parse_simple_selector(tokens, namespaces)
        if pseudo_element is not None or simple_selector is None:
            break
        simple_selectors.append(simple_selector)

    if simple_selectors or (type_selectors, pseudo_element) != (None, None):
        return CompoundSelector(simple_selectors), pseudo_element

    peek = tokens.peek()
    peek_type = peek.type if peek else "EOF"
    raise SelectorError(peek, f"expected a compound selector, got {peek_type}")


def parse_type_selector(tokens, namespaces):
    tokens.skip_whitespace()
    qualified_name = parse_qualified_name(tokens, namespaces)
    if qualified_name is None:
        return None

    simple_selectors = []
    namespace, local_name = qualified_name
    if local_name is not None:
        simple_selectors.append(LocalNameSelector(local_name))
    if namespace is not None:
        simple_selectors.append(NamespaceSelector(namespace))
    return simple_selectors


def parse_simple_selector(tokens, namespaces):
    peek = tokens.peek()
    if peek is None:
        return None, None
    if peek.type == "hash" and peek.is_identifier:
        tokens.next()
        return IDSelector(peek.value), None
    elif peek == ".":
        tokens.next()
        next = tokens.next()
        if next is None or next.type != "ident":
            raise SelectorError(next, f"Expected a class name, got {next}")
        return ClassSelector(next.value), None
    elif peek.type == "[] block":
        tokens.next()
        attr = parse_attribute_selector(TokenStream(peek.content), namespaces)
        return attr, None
    elif peek == ":":
        tokens.next()
        next = tokens.next()
        if next == ":":
            next = tokens.next()
            if next is None or next.type != "ident":
                raise SelectorError(next, f"Expected a pseudo-element name, got {next}")
            value = next.lower_value
            if value not in SUPPORTED_PSEUDO_ELEMENTS:
                raise SelectorError(
                    next, f"Expected a supported pseudo-element, got {value}"
                )
            return None, value
        elif next is not None and next.type == "ident":
            name = next.lower_value
            if name in ("before", "after", "first-line", "first-letter"):
                return None, name
            else:
                return PseudoClassSelector(name), None
        elif next is not None and next.type == "function":
            name = next.lower_name
            if name in ("is", "where", "not", "has"):
                return parse_logical_combination(next, namespaces, name), None
            else:
                return (FunctionalPseudoClassSelector(name, next.arguments), None)
        else:
            raise SelectorError(next, f"unexpected {next} token.")
    else:
        return None, None


def parse_logical_combination(matches_any_token, namespaces, name):
    forgiving = True
    relative = False
    if name == "is":
        selector_class = MatchesAnySelector
    elif name == "where":
        selector_class = SpecificityAdjustmentSelector
    elif name == "not":
        forgiving = False
        selector_class = NegationSelector
    elif name == "has":
        relative = True
        selector_class = RelationalSelector

    selectors = [
        selector
        for selector in parse(
            matches_any_token.arguments, namespaces, forgiving, relative
        )
        if selector.pseudo_element is None
    ]
    return selector_class(selectors)


def parse_attribute_selector(tokens, namespaces):
    tokens.skip_whitespace()
    qualified_name = parse_qualified_name(tokens, namespaces, is_attribute=True)
    if qualified_name is None:
        next = tokens.next()
        raise SelectorError(next, f"expected attribute name, got {next}")
    namespace, local_name = qualified_name

    tokens.skip_whitespace()
    peek = tokens.peek()
    if peek is None:
        operator = None
        value = None
    elif peek in ("=", "~=", "|=", "^=", "$=", "*="):
        operator = peek.value
        tokens.next()
        tokens.skip_whitespace()
        next = tokens.next()
        if next is None or next.type not in ("ident", "string"):
            next_type = "None" if next is None else next.type
            raise SelectorError(next, f"expected attribute value, got {next_type}")
        value = next.value
    else:
        raise SelectorError(peek, f"expected attribute selector operator, got {peek}")

    tokens.skip_whitespace()
    next = tokens.next()
    case_sensitive = None
    if next is not None:
        if next.type == "ident" and next.value.lower() == "i":
            case_sensitive = False
        elif next.type == "ident" and next.value.lower() == "s":
            case_sensitive = True
        else:
            raise SelectorError(next, f"expected ], got {next.type}")
    return AttributeSelector(namespace, local_name, operator, value, case_sensitive)


def parse_qualified_name(tokens, namespaces, is_attribute=False):
    """Return ``(namespace, local)`` for given tokens.

    Can also return ``None`` for a wildcard.

    The empty string for ``namespace`` means "no namespace".

    """
    peek = tokens.peek()
    if peek is None:
        return None
    if peek.type == "ident":
        first_ident = tokens.next()
        peek = tokens.peek()
        if peek != "|":
            namespace = "" if is_attribute else namespaces.get(None, None)
            return namespace, (first_ident.value, first_ident.lower_value)
        tokens.next()
        namespace = namespaces.get(first_ident.value)
        if namespace is None:
            raise SelectorError(
                first_ident, f"undefined namespace prefix: {first_ident.value}"
            )
    elif peek == "*":
        next = tokens.next()
        peek = tokens.peek()
        if peek != "|":
            if is_attribute:
                raise SelectorError(next, f"expected local name, got {next.type}")
            return namespaces.get(None, None), None
        tokens.next()
        namespace = None
    elif peek == "|":
        tokens.next()
        namespace = ""
    else:
        return None

    # If we get here, we just consumed '|' and set ``namespace``
    next = tokens.next()
    if next.type == "ident":
        return namespace, (next.value, next.lower_value)
    elif next == "*" and not is_attribute:
        return namespace, None
    else:
        raise SelectorError(next, f"expected local name, got {next.type}")


class SelectorError(ValueError):
    """A specialized ``ValueError`` for invalid selectors."""


class TokenStream:
    def __init__(self, tokens):
        self.tokens = iter(tokens)
        self.peeked = []  # In reversed order

    def next(self):
        if self.peeked:
            return self.peeked.pop()
        else:
            return next(self.tokens, None)

    def peek(self):
        if not self.peeked:
            self.peeked.append(next(self.tokens, None))
        return self.peeked[-1]

    def skip(self, skip_types):
        found = False
        while 1:
            peek = self.peek()
            if peek is None or peek.type not in skip_types:
                break
            self.next()
            found = True
        return found

    def skip_whitespace(self):
        return self.skip(["whitespace"])

    def skip_comment(self):
        return self.skip(["comment"])

    def skip_whitespace_and_comment(self):
        return self.skip(["comment", "whitespace"])


class Selector:
    def __init__(self, tree, pseudo_element=None):
        self.parsed_tree = tree
        self.pseudo_element = pseudo_element
        if pseudo_element is None:
            #: Tuple of 3 integers: http://www.w3.org/TR/selectors/#specificity
            self.specificity = tree.specificity
        else:
            a, b, c = tree.specificity
            self.specificity = a, b, c + 1

    def __repr__(self):
        pseudo = f"::{self.pseudo_element}" if self.pseudo_element else ""
        return f"{self.parsed_tree!r}{pseudo}"


class RelativeSelector:
    def __init__(self, combinator, selector):
        self.combinator = combinator
        self.selector = selector

    @property
    def specificity(self):
        return self.selector.specificity

    @property
    def pseudo_element(self):
        return self.selector.pseudo_element

    def __repr__(self):
        return (
            f"{self.selector!r}"
            if self.combinator == " "
            else f"{self.combinator} {self.selector!r}"
        )


class CombinedSelector:
    def __init__(self, left, combinator, right):
        #: Combined or compound selector
        self.left = left
        # One of `` `` (a single space), ``>``, ``+`` or ``~``.
        self.combinator = combinator
        #: compound selector
        self.right = right

    @property
    def specificity(self):
        a1, b1, c1 = self.left.specificity
        a2, b2, c2 = self.right.specificity
        return a1 + a2, b1 + b2, c1 + c2

    def __repr__(self):
        return f"{self.left!r}{self.combinator}{self.right!r}"


class CompoundSelector:
    def __init__(self, simple_selectors):
        self.simple_selectors = simple_selectors

    @property
    def specificity(self):
        if self.simple_selectors:
            # zip(*foo) turns [(a1, b1, c1), (a2, b2, c2), ...]
            # into [(a1, a2, ...), (b1, b2, ...), (c1, c2, ...)]
            return tuple(
                map(sum, zip(*(sel.specificity for sel in self.simple_selectors)))
            )
        else:
            return 0, 0, 0

    def __repr__(self):
        return "".join(map(repr, self.simple_selectors))


class LocalNameSelector:
    specificity = 0, 0, 1

    def __init__(self, local_name):
        self.local_name, self.lower_local_name = local_name

    def __repr__(self):
        return self.local_name


class NamespaceSelector:
    specificity = 0, 0, 0

    def __init__(self, namespace):
        #: The namespace URL as a string,
        #: or the empty string for elements not in any namespace.
        self.namespace = namespace

    def __repr__(self):
        if self.namespace == "":
            return "|"
        else:
            return f"{{{self.namespace}}}|"


class IDSelector:
    specificity = 1, 0, 0

    def __init__(self, ident):
        self.ident = ident

    def __repr__(self):
        return f"#{self.ident}"


class ClassSelector:
    specificity = 0, 1, 0

    def __init__(self, class_name):
        self.class_name = class_name

    def __repr__(self):
        return f".{self.class_name}"


class AttributeSelector:
    specificity = 0, 1, 0

    def __init__(self, namespace, name, operator, value, case_sensitive):
        self.namespace = namespace
        self.name, self.lower_name = name
        #: A string like ``=`` or ``~=``, or None for ``[attr]`` selectors
        self.operator = operator
        #: A string, or None for ``[attr]`` selectors
        self.value = value
        #: ``True`` if case-sensitive, ``False`` if case-insensitive, ``None``
        #: if depends on the document language
        self.case_sensitive = case_sensitive

    def __repr__(self):
        namespace = "*|" if self.namespace is None else f"{{{self.namespace}}}"
        case_sensitive = (
            ""
            if self.case_sensitive is None
            else f" {'s' if self.case_sensitive else 'i'}"
        )
        return f"[{namespace}{self.name}{self.operator}{self.value!r}{case_sensitive}]"


class PseudoClassSelector:
    specificity = 0, 1, 0

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return ":" + self.name


class FunctionalPseudoClassSelector:
    specificity = 0, 1, 0

    def __init__(self, name, arguments):
        self.name = name
        self.arguments = arguments

    def __repr__(self):
        return f":{self.name}{tuple(self.arguments)!r}"


class NegationSelector:
    def __init__(self, selector_list):
        self.selector_list = selector_list

    @property
    def specificity(self):
        if self.selector_list:
            return max(selector.specificity for selector in self.selector_list)
        else:
            return (0, 0, 0)

    def __repr__(self):
        return f":not({', '.join(repr(sel) for sel in self.selector_list)})"


class RelationalSelector:
    def __init__(self, selector_list):
        self.selector_list = selector_list

    @property
    def specificity(self):
        if self.selector_list:
            return max(selector.specificity for selector in self.selector_list)
        else:
            return (0, 0, 0)

    def __repr__(self):
        return f":has({', '.join(repr(sel) for sel in self.selector_list)})"


class MatchesAnySelector:
    def __init__(self, selector_list):
        self.selector_list = selector_list

    @property
    def specificity(self):
        if self.selector_list:
            return max(selector.specificity for selector in self.selector_list)
        else:
            return (0, 0, 0)

    def __repr__(self):
        return f":is({', '.join(repr(sel) for sel in self.selector_list)})"


class SpecificityAdjustmentSelector:
    def __init__(self, selector_list):
        self.selector_list = selector_list

    @property
    def specificity(self):
        return (0, 0, 0)

    def __repr__(self):
        return f":where({', '.join(repr(sel) for sel in self.selector_list)})"



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2022 Jonathan Neuhauser, jonathan.neuhauser@outlook.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""Join paths with lines or polygons"""

import itertools
from typing import List, Union

import inkex
from inkex.localization import inkex_gettext as _
from inkex.paths import ZoneClose, zoneClose, Line, Move, move
from inkex.turtle import PathGroupBuilder, PathBuilder


class Extrude(inkex.EffectExtension):
    """This effect draws lines between each nth node of each selected path.
    It can be chosen whether these regions are filled and whether the fill uses rectangles
    or copies of the path segments.
    The lines will be inserted between the two elements.
    """

    def add_arguments(self, pars):
        pars.add_argument("--tab")
        pars.add_argument(
            "-m",
            "--mode",
            default="lines",
            choices=["lines", "polygons", "snug"],
            help='Join paths with lines, polygons or copies of the segments ("snug")',
        )
        pars.add_argument(
            "-s",
            "--subpaths",
            default=True,
            type=inkex.Boolean,
            help="""If true, connecting lines will be inserted as subpaths of a single path.
                        If false, they will be inserted in newly created group. 
                        Only applies to mode=lines""",
        )

    @staticmethod
    def _handle_lines(manager, com1, com2):
        if not (
            isinstance(com1.command, (ZoneClose, zoneClose))
            or isinstance(com2.command, (ZoneClose, zoneClose))
        ):
            # For a closed subpath, the first line has already been drawn.
            manager.Move_to(*com1.end_point)
            manager.Line_to(*com2.end_point)

    @staticmethod
    def _handle_polygons(manager, com1, com2):
        if not (
            isinstance(com1.command, (Move, move))
            or isinstance(com2.command, (Move, move))
        ):
            # We skip if one of either commands is a "Move" command
            manager.Move_to(*com1.previous_end_point)
            for point in [
                com1.end_point,
                com2.end_point,
                com2.previous_end_point,
                com1.previous_end_point,
            ]:
                manager.Line_to(*point)

    @staticmethod
    def _handle_snug(manager, com1, com2):
        if not (
            isinstance(com1.command, (Move, move))
            or isinstance(com2.command, (Move, move))
        ):
            # We skip if one of either commands is a "Move" command
            manager.Move_to(*com1.previous_end_point)
            com1r = com1.command
            com2r = com2.reverse()
            doflag = True
            if isinstance(com1r, (ZoneClose, zoneClose)):
                # ZoneClose can not be used directly, must be converted to line
                com1r = Line(*com1.first_point)
                if com1.previous_end_point.is_close(com1.end_point):
                    doflag = False
            if doflag:
                manager.add([com1r, Line(*com2.end_point), com2r, ZoneClose()])

    def effect(self):
        paths: List[inkex.PathElement] = []
        for node in self.svg.selection.rendering_order().filter(inkex.ShapeElement):
            if isinstance(node, inkex.PathElement):
                node.apply_transform()
            paths.append(node)
        if len(paths) < 2:
            raise inkex.AbortExtension(_("Need at least 2 paths selected"))
        lines = self.options.mode.lower() == "lines"
        subpaths = self.options.subpaths and lines

        mode = (
            self._handle_lines
            if lines
            else (
                self._handle_polygons
                if self.options.mode.lower() == "polygons"
                else self._handle_snug
            )
        )

        if lines:
            style = {
                "fill": "none",
                "stroke": "#000000",
                "stroke-opacity": 1,
                "stroke-width": "1px",
            }
        else:
            style = {
                "fill": "#000000",
                "fill-opacity": 0.3,
                "stroke": "#000000",
                "stroke-opacity": 0.6,
                "stroke-width": "1px",
                "stroke-linejoin": "round",
            }

        for pa1, pa2 in itertools.combinations(paths, 2):
            manager = PathBuilder(style) if subpaths else PathGroupBuilder(style)
            for com1, com2 in zip(pa1.path.proxy_iterator(), pa2.path.proxy_iterator()):
                mode(manager, com1, com2)
                manager.terminate()
            manager.append_next(pa1)


if __name__ == "__main__":
    Extrude().run()



# SPDX-FileCopyrightText: 2022 Jonathan Neuhauser <jonathan.neuhauser@outlook.com>
# SPDX-FileCopyrightText: 2023 Software Freedom Conservancy <info@sfconservancy.org>
#
# SPDX-License-Identifier: GPL-2.0-or-later
import os
import sys

# definitions
HERE = os.path.abspath(os.path.dirname(__file__))
PARENT = os.path.abspath(os.path.dirname(HERE))

#
# We might get executed in HERE.
# This might lead to problems.
# In this case, we move one down.
#
import warnings

if not hasattr(warnings, "warn"):
    del sys.modules["warnings"]
    for i, path in reversed(list(enumerate(sys.path[:]))):
        if os.path.abspath(path) == HERE:
            del sys.path[i]

try:
    import inkex  # check inkex
    from . import svg  # check that we are a package
except ImportError:
    # This is suggested by https://docs.python-guide.org/writing/structure/.
    sys.path.insert(0, PARENT)
    # raise ImportError(f"{HERE}, {PARENT}")
    try:
        from inkvn.vninput import main
    except ImportError:
        raise
else:
    # All modules are installed and we are in a package
    # seems like we should be executed.
    from .vninput import main

main()



# coding=utf-8
#
# Copyright (C) 2018 Martin Owens
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
#
"""
SVG specific utilities for tests.
"""

from lxml import etree

from inkex import SVG_PARSER


def svg(svg_attrs=""):
    """Returns xml etree based on a simple SVG element.

    svg_attrs: A string containing attributes to add to the
        root <svg> element of a minimal SVG document.
    """
    return etree.fromstring(
        str.encode(
            '<?xml version="1.0" encoding="UTF-8" standalone="no"?>'
            f"<svg {svg_attrs}></svg>"
        ),
        parser=SVG_PARSER,
    )


def svg_unit_scaled(width_unit):
    """Same as svg, but takes a width unit (top-level transform) for the new document.

    The transform is the ratio between the SVG width and the viewBox width.
    """
    return svg(f'width="1{width_unit}" viewBox="0 0 1 1"')


def svg_file(filename):
    """Parse an svg file and return it's document root"""
    with open(filename, "r", encoding="utf-8") as fhl:
        doc = etree.parse(fhl, parser=SVG_PARSER)
        return doc.getroot()



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2005 Aaron Spike, aaron@ekips.org
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""Adds a "drop shadow" to any selected number of path objects. Optionally, the stroke
color can be used for the shadow"""

import math

import inkex
from inkex.paths import (
    Move,
    Line,
    Curve,
    ZoneClose,
    Arc,
    Path,
    Vert,
    Horz,
    TepidQuadratic,
    Quadratic,
    Smooth,
)
from inkex.transforms import Vector2d
from inkex.bezier import beziertatslope, beziersplitatt


class LongShadow(inkex.EffectExtension):
    """Generate a motion path"""

    def add_arguments(self, pars):
        pars.add_argument(
            "-a",
            "--angle",
            type=float,
            default=45.0,
            help="direction of the motion vector",
        )
        pars.add_argument(
            "-m",
            "--magnitude",
            type=float,
            default=100.0,
            help="magnitude of the motion vector",
        )
        pars.add_argument(
            "-f",
            "--fillwithstroke",
            type=inkex.Boolean,
            default=False,
            help="fill shadow with stroke color if set",
        )

    @staticmethod
    def makeface(last, segment, facegroup, delx, dely):
        """translate path segment along vector"""
        elem = facegroup.add(inkex.PathElement())

        npt = segment.translate([delx, dely])

        # reverse direction of path segment
        if isinstance(segment, Curve):
            rev = Curve(npt.x3, npt.y3, npt.x2, npt.y2, last[0] + delx, last[1] + dely)
        elif isinstance(segment, Line):
            rev = Line(last[0] + delx, last[1] + dely)
        else:
            raise RuntimeError("Unexpected segment type {}".format(type(segment)))

        elem.path = inkex.Path(
            [
                Move(last[0], last[1]),
                segment,
                npt.to_line(Vector2d()),
                rev,
                ZoneClose(),
            ]
        )

    def effect(self):
        delx = math.cos(math.radians(self.options.angle)) * self.options.magnitude
        dely = math.sin(math.radians(self.options.angle)) * self.options.magnitude
        for node in self.svg.selection.filter_nonzero(inkex.PathElement):
            group = node.getparent().add(inkex.Group())
            facegroup = group.add(inkex.Group())
            group.append(node)

            # we want delx and dely values to be correct even for the transformed path,
            # so transform them back, but ignore the translate of the transform
            trans = -node.transform
            local_delx = trans.a * delx + trans.c * dely
            local_dely = trans.b * delx + trans.d * dely

            if node.transform:
                group.transform = node.transform
                node.transform = None

            facegroup.style = node.style
            if self.options.fillwithstroke:
                stroke = facegroup.style("stroke")
                if stroke is not None and isinstance(stroke, inkex.Color):
                    facegroup.style["fill"] = stroke
                    facegroup.style["fill-opacity"] = facegroup.style("stroke-opacity")
            reset_origin = True
            for cmd_proxy in node.path.to_absolute().proxy_iterator():
                # for each subpath, reset the origin of the following computations to the first
                # node of the subpath -> i.e. after a Z command, move the origin to the end point
                # of the next command
                if reset_origin:
                    first_point = cmd_proxy.end_point
                    reset_origin = False
                if isinstance(cmd_proxy.command, ZoneClose):
                    reset_origin = True
                self.process_segment(
                    cmd_proxy, facegroup, local_delx, local_dely, first_point
                )

    @staticmethod
    def process_segment(cmd_proxy, facegroup, delx, dely, first_point):
        """Process each segments"""

        segments = []
        if isinstance(
            cmd_proxy.command, (Curve, Smooth, TepidQuadratic, Quadratic, Arc)
        ):
            prev = cmd_proxy.previous_end_point
            for curve in cmd_proxy.to_curves():
                bez = [prev] + curve.to_bez()
                prev = curve.end_point(cmd_proxy.first_point, prev)
                tees = [t for t in beziertatslope(bez, (dely, delx)) if 0 < t < 1]
                tees.sort()
                if len(tees) == 1:
                    one, two = beziersplitatt(bez, tees[0])
                    segments.append(Curve(*(one[1] + one[2] + one[3])))
                    segments.append(Curve(*(two[1] + two[2] + two[3])))
                elif len(tees) == 2:
                    one, two = beziersplitatt(bez, tees[0])
                    two, three = beziersplitatt(two, tees[1])
                    segments.append(Curve(*(one[1] + one[2] + one[3])))
                    segments.append(Curve(*(two[1] + two[2] + two[3])))
                    segments.append(Curve(*(three[1] + three[2] + three[3])))
                else:
                    segments.append(curve)
        elif isinstance(cmd_proxy.command, (Line, Curve)):
            segments.append(cmd_proxy.command)
        elif isinstance(cmd_proxy.command, ZoneClose):
            segments.append(Line(*first_point))
        elif isinstance(cmd_proxy.command, (Vert, Horz)):
            segments.append(cmd_proxy.command.to_line(cmd_proxy.end_point))

        for seg in Path(
            [Move(*cmd_proxy.previous_end_point)] + segments
        ).proxy_iterator():
            if isinstance(seg.command, Move):
                continue
            LongShadow.makeface(
                seg.previous_end_point, seg.command, facegroup, delx, dely
            )


if __name__ == "__main__":
    LongShadow().run()



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2005,2007 Aaron Spike, aaron@ekips.org
# Copyright (C) 2009 Alvin Penner, penner@vaxxine.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
This extension converts a path into a dashed line using 'stroke-dasharray'
It is a modification of the file addnodes.py
"""

import inkex
from inkex import bezier, CubicSuperPath, Group, PathElement
from inkex.localization import inkex_gettext as _
from inkex.paths.interfaces import ILengthSettings


class Dashit(inkex.EffectExtension):
    """Extension to convert paths into dash-array line"""

    def __init__(self):
        super(Dashit, self).__init__()
        self.not_converted = []

    def effect(self):
        for node in self.svg.selection:
            self.convert2dash(node)
        if self.not_converted:
            inkex.errormsg(
                _("Total number of objects not converted: {}\n").format(
                    len(self.not_converted)
                )
            )
            # return list of IDs in case the user needs to find a specific object
            inkex.utils.debug(self.not_converted)

    def convert2dash(self, node):
        """Convert each selected node's dash array"""
        if isinstance(node, Group):
            for child in node:
                self.convert2dash(child)
        elif isinstance(node, PathElement):
            self._convert(node)
        else:
            self.not_converted.append(node.get("id"))

    @staticmethod
    def _convert(node):
        dashes = []
        offset = 0
        overlap = 0
        dashes = node.get_computed_style("stroke-dasharray")
        offset = float(node.get_computed_style("stroke-dashoffset"))
        # Correct negative offsets
        while offset < 0:
            offset += sum(dashes)
        if not dashes:
            return
        new = inkex.Path()
        segment: inkex.Path.PathCommandProxy
        for segment in node.path.to_absolute().proxy_iterator():
            ismove = segment.letter == "M"
            if ismove:
                # Start a new subpath, reset dash counter.
                idash = 0
                dash = dashes[0]
                remaining_length = offset
                firstdrawn_index = -1
            else:
                remaining_length = segment.length() + overlap
                current = segment
            while dash < remaining_length:
                if not ismove:
                    first_part, current = current.split(current.ilength(dash - overlap))
                    if idash % 2:  # create a gap
                        new.append(inkex.paths.Move(first_part.cend_point))
                    else:
                        if firstdrawn_index == -1:
                            firstdrawn_index = len(new)
                        new.append(first_part.command)
                remaining_length = remaining_length - dash
                idash = (idash + 1) % len(dashes)
                dash = dashes[idash]
                overlap = 0
                # We have already drawn a gap
                if firstdrawn_index == -1:
                    firstdrawn_index = None
            if ismove:
                new.append(segment.command)
            else:
                if idash % 2:  # Process the final part of the segment
                    new.append(inkex.paths.Move(current.cend_point))
                else:
                    if current.letter == "Z":
                        # In case of ZoneClose:
                        # Replace the firstdrawn index with a moveto, and append the
                        # command to the end to fix linejoins
                        new.append(inkex.paths.Line(current.cend_point))
                        if firstdrawn_index is not None and firstdrawn_index > -1:
                            index = firstdrawn_index
                            while new[index].letter != "M":
                                element = new[index]
                                new[index] = inkex.paths.Move(
                                    element.cend_point(0j, 0j)
                                )
                                new.append(element)
                                index += 1
                    else:
                        if firstdrawn_index == -1:
                            firstdrawn_index = len(new)
                        new.append(current.to_non_shorthand())

            overlap = remaining_length
        node.style.pop("stroke-dasharray")
        node.pop("sodipodi:type")
        node.path = new


if __name__ == "__main__":
    Dashit().run()



"""
VNGroupElement
"""

from dataclasses import dataclass
from typing import List

from .base import VNBaseElement


@dataclass
class VNGroupElement(VNBaseElement):
    """Group Element properties."""

    groupElements: List[VNBaseElement]



# coding=utf-8
#
# Copyright (C) 2023 - Jonathan Neuhauser <jonathan.neuhauser@outlook.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""CSS evaluation logic, forked from cssselect2 (rewritten without eval, targeted to
our data structure). CSS selectors are compiled into boolean evaluator functions.
All HTML-specific code has been removed, and we don't duplicate the tree data structure
but work on the normal tree."""

import re
from lxml import etree
from typing import Union, List
from tinycss2.nth import parse_nth

from . import parser
from .parser import SelectorError

# http://dev.w3.org/csswg/selectors/#whitespace
split_whitespace = re.compile("[^ \t\r\n\f]+").findall


def ascii_lower(string):  # from webencodings
    r"""Transform (only) ASCII letters to lower case: A-Z is mapped to a-z."""
    return string.encode("utf8").lower().decode("utf8")


# pylint: disable=protected-access,comparison-with-callable,invalid-name,bad-super-call
# pylint: disable=unnecessary-lambda-assignment


## Iterators without comments.
def iterancestors(element):
    """Iterate over ancestors but ignore comments."""
    for e in element.iterancestors():
        if isinstance(e, etree._Comment):
            continue
        yield e


def iterdescendants(element):
    """Iterate over descendants but ignore comments"""
    for e in element.iterdescendants():
        if isinstance(e, etree._Comment):
            continue
        yield e


def itersiblings(element, preceding=False):
    """Iterate over descendants but ignore comments"""
    for e in element.itersiblings(preceding=preceding):
        if isinstance(e, etree._Comment):
            continue
        yield e


def iterchildren(element):
    """Iterate over children but ignore comments"""
    for e in element.iterchildren():
        if isinstance(e, etree._Comment):
            continue
        yield e


def getprevious(element):
    """Get the previous non-comment element"""
    for e in itersiblings(element, preceding=True):
        return e
    return None


def getnext(element):
    """Get the next non-comment element"""
    for e in itersiblings(element, preceding=False):
        return e
    return None


def FALSE(_el):
    """Always returns 0"""
    return 0


def TRUE(_el):
    """Always returns 1"""
    return 1


class BooleanCompiler:
    def __init__(self) -> None:
        self._func_map = {
            parser.CombinedSelector: self._compile_combined,
            parser.CompoundSelector: self._compile_compound,
            parser.NegationSelector: self._compile_negation,
            parser.RelationalSelector: self._compile_relational,
            parser.MatchesAnySelector: self._compile_any,
            parser.SpecificityAdjustmentSelector: self._compile_any,
            parser.LocalNameSelector: self._compile_local_name,
            parser.NamespaceSelector: self._compile_namespace,
            parser.ClassSelector: self._compile_class,
            parser.IDSelector: self._compile_id,
            parser.AttributeSelector: self._compile_attribute,
            parser.PseudoClassSelector: self._compile_pseudoclass,
            parser.FunctionalPseudoClassSelector: self._compile_functional_pseudoclass,
        }

    def _compile_combined(self, selector: parser.CombinedSelector):
        left_inside = self.compile_node(selector.left)
        if left_inside == FALSE:
            return FALSE  # 0 and x == 0
        if left_inside == TRUE:
            # 1 and x == x, but the element matching 1 still needs to exist.
            if selector.combinator in (" ", ">"):
                left = lambda el: el.getparent() is not None

            elif selector.combinator in ("~", "+"):
                left = lambda el: getprevious(el) is not None

            else:
                raise SelectorError("Unknown combinator", selector.combinator)
        elif selector.combinator == " ":
            left = lambda el: any((left_inside(e)) for e in el.ancestors())

        elif selector.combinator == ">":
            left = lambda el: el.getparent() is not None and left_inside(el.getparent())

        elif selector.combinator == "+":
            left = lambda el: getprevious(el) is not None and left_inside(
                getprevious(el)
            )

        elif selector.combinator == "~":
            left = lambda el: any(
                (left_inside(e)) for e in itersiblings(el, preceding=True)
            )

        else:
            raise SelectorError("Unknown combinator", selector.combinator)

        right = self.compile_node(selector.right)
        if right == FALSE:
            return FALSE  # 0 and x == 0
        if right == TRUE:
            return left  # 1 and x == x
        # Evaluate combinators right to left
        return lambda el: right(el) and left(el)

    def _compile_compound(self, selector: parser.CompoundSelector):
        sub_expressions = [
            expr
            for expr in map(self.compile_node, selector.simple_selectors)
            if expr != TRUE
        ]
        if len(sub_expressions) == 1:
            return sub_expressions[0]
        if FALSE in sub_expressions:
            return FALSE
        if sub_expressions:
            return lambda e: all(expr(e) for expr in sub_expressions)
        return TRUE  # all([]) == True

    def _compile_negation(self, selector: parser.NegationSelector):
        sub_expressions = [
            expr
            for expr in [
                self.compile_node(selector.parsed_tree)
                for selector in selector.selector_list
            ]
            if expr != TRUE
        ]
        if not sub_expressions:
            return FALSE
        return lambda el: not any(expr(el) for expr in sub_expressions)

    @staticmethod
    def _get_subexpr(expression, relative_selector):
        """Helper function for RelationalSelector"""
        if relative_selector.combinator == " ":
            return lambda el: any(expression(e) for e in iterdescendants(el))
        if relative_selector.combinator == ">":
            return lambda el: any(expression(e) for e in iterchildren(el))
        if relative_selector.combinator == "+":
            return lambda el: expression(next(itersiblings(el)))
        if relative_selector.combinator == "~":
            return lambda el: any(expression(e) for e in itersiblings(el))
        raise SelectorError(
            f"Unknown relational selector '{relative_selector.combinator}'"
        )

    def _compile_relational(self, selector: parser.RelationalSelector):
        sub_expr = []

        for relative_selector in selector.selector_list:
            expression = self.compile_node(relative_selector.selector.parsed_tree)
            if expression == FALSE:
                continue
            sub_expr.append(self._get_subexpr(expression, relative_selector))
        return lambda el: any(expr(el) for expr in sub_expr)

    def _compile_any(
        self,
        selector: Union[
            parser.MatchesAnySelector, parser.SpecificityAdjustmentSelector
        ],
    ):
        sub_expressions = [
            expr
            for expr in [
                self.compile_node(selector.parsed_tree)
                for selector in selector.selector_list
            ]
            if expr != FALSE
        ]
        if not sub_expressions:
            return FALSE
        return lambda el: any(expr(el) for expr in sub_expressions)

    def _compile_local_name(self, selector: parser.LocalNameSelector):
        return lambda el: el.TAG == selector.local_name

    def _compile_namespace(self, selector: parser.NamespaceSelector):
        return lambda el: el.NAMESPACE == selector.namespace

    def _compile_class(self, selector: parser.ClassSelector):
        return lambda el: selector.class_name in el.classes

    def _compile_id(self, selector: parser.IDSelector):
        return lambda el: super(etree.ElementBase, el).get("id", None) == selector.ident  # type: ignore

    def _compile_attribute(self, selector: parser.AttributeSelector):
        if selector.namespace is not None:
            if selector.namespace:
                key_func = lambda el: (
                    f"{{{selector.namespace}}}{selector.name}"
                    if el.NAMESPACE != selector.namespace
                    else selector.name
                )

            else:
                key_func = lambda el: selector.name

            value = selector.value
            if selector.case_sensitive is False:
                value = value.lower()

                attribute_value = (
                    lambda el: super(etree.ElementBase, el)
                    .get(key_func(el), "")  # type: ignore
                    .lower()
                )

            else:
                attribute_value = lambda el: super(etree.ElementBase, el).get(  # type: ignore
                    key_func(el), ""
                )

            if selector.operator is None:
                return lambda el: key_func(el) in el.attrib
            if selector.operator == "=":
                return lambda el: (
                    key_func(el) in el.attrib and attribute_value(el) == value
                )
            if selector.operator == "~=":
                return (
                    FALSE
                    if len(value.split()) != 1 or value.strip() != value
                    else lambda el: value in split_whitespace(attribute_value(el))
                )
            if selector.operator == "|=":
                return lambda el: (
                    key_func(el) in el.attrib
                    and (
                        attribute_value(el) == value
                        or attribute_value(el).startswith(value + "-")
                    )
                )
            if selector.operator == "^=":
                if value:
                    return lambda el: attribute_value(el).startswith(value)
                return FALSE
            if selector.operator == "$=":
                return (
                    (lambda el: attribute_value(el).endswith(value)) if value else FALSE
                )
            if selector.operator == "*=":
                return (lambda el: value in attribute_value(el)) if value else FALSE
            raise SelectorError("Unknown attribute operator", selector.operator)
        # In any namespace
        raise NotImplementedError  # TODO

    def _compile_pseudoclass(self, selector: parser.PseudoClassSelector):
        if selector.name in ("link", "any-link", "local-link"):

            def ancestors_or_self(el):
                yield el
                yield from iterancestors(el)

            return lambda el: any(
                e.TAG == "a" and super(etree.ElementBase, e).get("href", "") != ""  # type: ignore
                for e in ancestors_or_self(el)
            )
        if selector.name in (
            "visited",
            "hover",
            "active",
            "focus",
            "focus-within",
            "focus-visible",
            "target",
            "target-within",
            "current",
            "past",
            "future",
            "playing",
            "paused",
            "seeking",
            "buffering",
            "stalled",
            "muted",
            "volume-locked",
            "user-valid",
            "user-invalid",
        ):
            # Not applicable in a static context: never match.
            return FALSE
        if selector.name in ("enabled", "disabled", "checked"):
            # Not applicable to SVG
            return FALSE
        if selector.name in ("root", "scope"):
            return lambda el: el.getparent() is None
        if selector.name == "first-child":
            return lambda el: getprevious(el) is None
        if selector.name == "last-child":
            return lambda el: getnext(el) is None
        if selector.name == "first-of-type":
            return lambda el: all(
                s.tag != el.tag for s in itersiblings(el, preceding=True)
            )
        if selector.name == "last-of-type":
            return lambda el: all(s.tag != el.tag for s in itersiblings(el))
        if selector.name == "only-child":
            return lambda el: getnext(el) is None and getprevious(el) is None
        if selector.name == "only-of-type":
            return lambda el: all(s.tag != el.tag for s in itersiblings(el)) and all(
                s.tag != el.tag for s in itersiblings(el, preceding=True)
            )
        if selector.name == "empty":
            return lambda el: not list(el) and el.text is None
        raise SelectorError("Unknown pseudo-class", selector.name)

    def _compile_lang(self, selector: parser.FunctionalPseudoClassSelector):
        langs = []
        tokens = [
            token
            for token in selector.arguments
            if token.type not in ("whitespace", "comment")
        ]
        while tokens:
            token = tokens.pop(0)
            if token.type == "ident":
                langs.append(token.lower_value)
            elif token.type == "string":
                langs.append(ascii_lower(token.value))
            else:
                raise SelectorError("Invalid arguments for :lang()")
            if tokens:
                token = tokens.pop(0)
                if token.type != "ident" and token.value != ",":
                    raise SelectorError("Invalid arguments for :lang()")

        def haslang(el, lang):
            print(
                el.get("lang"),
                lang,
                el.get("lang", "") == lang or el.get("lang", "").startswith(lang + "-"),
            )
            return el.get("lang", "").lower() == lang or el.get(
                "lang", ""
            ).lower().startswith(lang + "-")

        return lambda el: any(
            haslang(el, lang) or any(haslang(el2, lang) for el2 in iterancestors(el))
            for lang in langs
        )

    def _compile_functional_pseudoclass(
        self, selector: parser.FunctionalPseudoClassSelector
    ):
        if selector.name == "lang":
            return self._compile_lang(selector)
        nth: List[str] = []
        selector_list: List[str] = []
        current_list = nth
        for argument in selector.arguments:
            if argument.type == "ident" and argument.value == "of":
                if current_list is nth:
                    current_list = selector_list
                    continue
            current_list.append(argument)

        if selector_list:
            compiled = tuple(
                self.compile_node(selector.parsed_tree)
                for selector in parser.parse(selector_list)
            )
            test = lambda el: all(expr(el) for expr in compiled)

        else:
            test = TRUE

        if selector.name == "nth-child":
            count = lambda el: sum(
                1 for e in itersiblings(el, preceding=True) if test(e)
            )

        elif selector.name == "nth-last-child":
            count = lambda el: sum(1 for e in itersiblings(el) if test(e))
        elif selector.name == "nth-of-type":
            count = lambda el: sum(
                1
                for s in (e for e in itersiblings(el, preceding=True) if test(e))
                if s.tag == el.tag
            )

        elif selector.name == "nth-last-of-type":
            count = lambda el: sum(
                1 for s in (e for e in itersiblings(el) if test(e)) if s.tag == el.tag
            )

        else:
            raise SelectorError("Unknown pseudo-class", selector.name)

        count_func = lambda el: count(el) if test(el) else float("nan")

        result = parse_nth(nth)
        if result is None:
            raise SelectorError(f"Invalid arguments for :{selector.name}()")
        a, b = result
        # x is the number of siblings before/after the element
        # Matches if a positive or zero integer n exists so that:
        # x = a*n + b-1
        # x = a*n + B
        B = b - 1
        if a == 0:
            # x = B
            return lambda el: count_func(el) == B

        # n = (x - B) / a
        def evaluator(el):
            n, r = divmod(count_func(el) - B, a)
            return r == 0 and n >= 0

        return evaluator

    def compile_node(self, selector):
        """Return a boolean expression, as a callable.

        When evaluated in a context where the `el` variable is an
        :class:`cssselect2.tree.Element` object, tells whether the element is a
        subject of `selector`.

        """
        try:
            return self._func_map[selector.__class__](selector)
        except KeyError as e:
            raise TypeError(type(selector), selector) from e


CSSCompiler = BooleanCompiler()



#!/usr/bin/env python3
# coding=utf-8
# coding=utf-8
#
# Copyright (C) 2006 Aaron Spike, aaron@ekips.org
# Copyright (C) 2010 Nicolas Dufour, nicoduf@yahoo.fr (color options)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import inkex
from inkex.localization import inkex_gettext as _

MARKERS = ["marker", "marker-start", "marker-mid", "marker-end"]


class ColorMarkers(inkex.EffectExtension):
    """Add marker stroke to outline markers on selected objects."""

    def add_arguments(self, pars):
        pars.add_argument(
            "--modify",
            type=inkex.Boolean,
            default=False,
            help="Do not create a copy, modify the markers",
        )
        pars.add_argument(
            "--type",
            dest="fill_type",
            default="solid",
            help="Replace the markers' fill with the object stroke or fill color",
        )
        pars.add_argument(
            "--alpha",
            type=inkex.Boolean,
            dest="assign_alpha",
            default=True,
            help="Assign the object fill and stroke alpha to the markers",
        )
        pars.add_argument(
            "--invert",
            type=inkex.Boolean,
            default=False,
            help="Invert fill and stroke colors",
        )
        pars.add_argument(
            "--assign_fill",
            type=inkex.Boolean,
            default=True,
            help="Assign a fill color to the markers",
        )
        pars.add_argument(
            "--fill_color",
            type=inkex.Color,
            default=inkex.Color(1364325887),
            help="Choose a custom fill color",
        )
        pars.add_argument(
            "--assign_stroke",
            type=inkex.Boolean,
            default=True,
            help="Assign a stroke color to the markers",
        )
        pars.add_argument(
            "--stroke_color",
            type=inkex.Color,
            default=inkex.Color(1364325887),
            help="Choose a custom fill color",
        )
        pars.add_argument(
            "--tab",
            type=self.arg_method("method"),
            default=self.method_custom,
            help="The selected UI-tab when OK was pressed",
        )
        pars.add_argument(
            "--colortab", help="The selected custom color tab when OK was pressed"
        )

    def method_custom(self, _):
        """Choose custom colors"""
        fill = self.options.fill_color if self.options.assign_fill else None
        stroke = self.options.stroke_color if self.options.assign_stroke else None
        return fill, stroke

    def method_object(self, style):
        """Use object colors"""
        fill = style.get_color("fill")
        stroke = style.get_color("stroke")

        if self.options.fill_type == "solid":
            fill = stroke
        elif self.options.fill_type == "filled":
            stroke = None
        elif self.options.invert:
            fill, stroke = stroke, fill

        if not self.options.assign_alpha:
            # Remove alpha values
            fill = fill.to_rgb()
            stroke = stroke.to_rgb()

        return fill, stroke

    def effect(self):
        for node in self.svg.selection.values():
            fill, stroke = self.options.tab(node.style)

            for attr in MARKERS:
                marker_node = node.style(attr)
                if not isinstance(marker_node, inkex.Marker):
                    continue

                if marker_node is None and attr in node.style:
                    inkex.errormsg(_("unable to locate marker: %s") % node.style[attr])
                    continue

                marker_id = marker_node.get_id()

                if not self.options.modify:
                    marker_node = marker_node.copy()
                    self.svg.defs.append(marker_node)
                    marker_id = self.svg.get_unique_id(marker_id)
                marker_node.set("id", marker_id)
                marker_node.set("inkscape:stockid", marker_id)

                node.style[attr] = marker_node

                for child in marker_node:
                    if stroke is not None:
                        child.style.set_color(stroke, "stroke")
                    if fill is not None:
                        child.style.set_color(fill, "fill")


if __name__ == "__main__":
    ColorMarkers().run()



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2005 Aaron Spike, aaron@ekips.org
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
import math

import inkex
from inkex import TextElement, Circle
from inkex.localization import inkex_gettext as _


class NumberNodes(inkex.EffectExtension):
    """Replace the selection's nodes with numbered dots according to the options"""

    def add_arguments(self, pars):
        pars.add_argument(
            "--dotsize", default="10px", help="Size of the dots on the path nodes"
        )
        pars.add_argument("--fontsize", default="20px", help="Size of node labels")
        pars.add_argument(
            "--start", type=int, default=1, help="First number in the sequence"
        )
        pars.add_argument(
            "--step", type=int, default=1, help="Numbering step between two nodes"
        )
        pars.add_argument("--tab", help="The selected UI-tab when OK was pressed")

    def effect(self):
        filtered = self.svg.selection.filter(inkex.PathElement)
        if not filtered:
            raise inkex.AbortExtension(_("Please select at least one path object."))
        for node in filtered:
            self.add_dot(node)

    def add_dot(self, node: inkex.PathElement):
        """Add a dot label for this path element"""
        group: inkex.Group = node.getparent().add(inkex.Group())
        dot_group = group.add(inkex.Group())
        num_group = group.add(inkex.Group())
        path_trans_applied = node.path.transform(node.composed_transform())
        group.transform = -node.getparent().composed_transform()

        style = inkex.Style({"stroke": "none", "fill": "#000"})

        for step, (x, y) in enumerate(path_trans_applied.end_points):
            circle = dot_group.add(
                Circle(
                    cx=str(x),
                    cy=str(y),
                    r=str(self.svg.unittouu(self.options.dotsize) / 2),
                )
            )
            circle.style = style
            num_group.append(
                self.add_text(
                    x + (self.svg.unittouu(self.options.dotsize) / 2),
                    y - (self.svg.unittouu(self.options.dotsize) / 2),
                    self.options.start + (self.options.step * step),
                )
            )

        node.delete()

    def add_text(self, x, y, text):
        """Add a text label at the given location"""
        elem = TextElement(x=str(x), y=str(y))
        elem.text = str(text)
        elem.style = {
            "font-size": self.svg.unittouu(self.options.fontsize),
            "fill-opacity": "1.0",
            "stroke": "none",
            "font-weight": "normal",
            "font-style": "normal",
            "fill": "#999",
        }
        return elem


if __name__ == "__main__":
    NumberNodes().run()



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2022 Jonathan Neuhauser (jonathan.neuhauser@outlook.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#


"""HPGL Input state machine"""

import math
from typing import List, Optional, Tuple, Dict, Union
from pyparsing import ParseResults
import inkex


class ListWithCallback(list):
    """A list that modifies elements with a callback before appending them"""

    def __init__(self, callback):
        super().__init__()
        self.callback = callback

    def append(self, item):
        super().append(self.callback(item))


PathData = ListWithCallback


class HPGLStateMachine:
    """This a HPGL plotter"""

    def __init__(self, root: inkex.Layer, options):
        self.options = options
        self.x: float
        self.y: float
        self.pendown: bool
        self.current_path: PathData
        self.current_parent: inkex.Layer
        self.absolute_mode: bool
        self.root = root
        self.polygon_mode = False
        self.polygon_buffer: List[PathData] = []
        # Start positions of polygon modes
        self.startx = 0
        self.starty = 0
        self.transform_manager = TransformManager(options)
        self.style_manager = StyleManager(options)
        self.initialize()

    def create_group(self):
        """Manage group"""
        # We'll set transforms; if the current layer is empty, don't create a new one
        self.finalize_path()
        if hasattr(self, "current_parent") and len(self.current_parent) == 0:
            group = self.current_parent
        else:
            group = inkex.Group()
            self.root.add(group)
            self.current_parent = group

        group.transform, clip = self.transform_manager.get_transform_and_clip(
            self.root.transform
        )
        if clip is not None:
            self.root.root.defs.append(clip)
            group.clip = clip
        self.initialize_current_path()
        self.current_path.append(inkex.paths.Move(self.x, self.y))

    def df_command(self):
        """default values"""
        self.style_manager.default_values()
        self.transform_manager.default_values()
        self.polygon_buffer = []
        self.initialize_current_path()
        self.create_group()  # in case scaling was turned off

    def initialize(self):
        """Reset state to default"""
        self.transform_manager.initialize()
        self.style_manager.initialize()
        self.x = 0
        self.y = 0
        self.pendown = False
        self.absolute_mode = True
        self.df_command()
        self.create_group()

    def initialize_current_path(self):
        """Initialize the current path with a custom append method"""

        def callback(new_elem):
            # we store the start position additionally to the path element itself
            result = [new_elem, self.pendown, self.x, self.y]
            if new_elem.is_absolute:
                self.x, self.y = new_elem.args[-2:]
            else:
                self.x, self.y = self.x + new_elem.args[-2], self.y + new_elem.args[-1]
            return result

        self.current_path = ListWithCallback(callback)

    def append_path_data(self, cmd, x, y):
        """Given lists of coordinates x, y, adds those to the path with command cmd,
        and sets the pen's position as the last of those"""

    def _line_command(self, key, xvals=None, yvals=None):
        if key in ["PA", "PR"]:
            self.absolute_mode = key == "PA"
        if key in ["PD", "PU"]:
            self.pendown = key == "PD"
        cmd = inkex.paths.Line if self.absolute_mode else inkex.paths.line
        if xvals is not None and yvals is not None:
            for xi, yi in zip(xvals, yvals):
                self.current_path.append(cmd(xi, yi))

    def line_command(self, vals: ParseResults):
        """Handle pu, pd, pa, pr"""
        vals_dict = vals.as_dict()
        if "X" in vals_dict and "Y" in vals_dict:
            self._line_command(vals_dict["key"], vals_dict["X"], vals_dict["Y"])
        else:
            self._line_command(vals_dict["key"])

    @staticmethod
    def create_path(data: List[PathData], fill=False, pmode=False) -> inkex.Path:
        """Create a path out of the stored points"""
        result = inkex.Path()

        for subpath in data:
            for j, (cmd, pendown, xstart, ystart) in enumerate(subpath):
                # For simplicity, commands are converted to absolute in polygon mode
                drawn_before = any(subpath[k][1] for k in range(j))
                if pmode:
                    cmd = cmd.to_absolute(inkex.Vector2d(xstart, ystart))
                    if not fill and pendown and not subpath[j - 1][1] and drawn_before:
                        # When edging a polygon, a line is drawn to "skip" (possibly
                        # subsequent) pen-up commands (but
                        # only if there already was a drawn command in this subpolygon)
                        result.append(inkex.paths.Line(xstart, ystart))
                # Don't draw commands when pen was up
                if not (not pendown and pmode and not fill and drawn_before):
                    # Also the first command of each subpath in polygon mode is
                    # converted: "When you use (PM1), the point after (PM1) becomes the
                    # first point of the next subpolygon. This move is not used as a
                    # boundary when filling a polygon with FP."
                    if (j == 0 and pmode) or (not pmode and not pendown):
                        repl = inkex.paths.move if cmd.is_relative else inkex.paths.Move
                        result.append(repl(*cmd.args[-2:]))
                    else:
                        result.append(cmd)
                if pmode and j == len(subpath) - 1:
                    result.append(inkex.paths.ZoneClose())
        # Ensure that the first path command is a moveto
        if result and result[0].letter.lower() != "m":
            result.insert(0, inkex.paths.move(0, 0))
        # Remove unnecessary moveto in the beginning of a path
        for i, cmd in enumerate(result.proxy_iterator()):
            if cmd.command.letter.lower() != "m":
                result = inkex.Path([inkex.paths.Move(*cmd.first_point)] + result[i:])
                break
        return result

    def finalize_path(self) -> None:
        """Append the current path to the parent if it's effectively non-empty"""
        result = self.create_path([self.current_path], False, False)
        if self.options.break_apart:
            result_list = result.break_apart()
        else:
            result_list = [result]
        for res in result_list:
            if any(i.letter not in "mM" for i in res):
                if self.current_path not in self.polygon_buffer:
                    pel = inkex.PathElement()
                    pel.path = res
                    self.current_parent.add(pel)
                    self.style_manager.set_stroke(
                        pel, p1=self.transform_manager.p1, p2=self.transform_manager.p2
                    )

        self.initialize_current_path()
        self.current_path.append(inkex.paths.Move(self.x, self.y))

    def ci_command(self, vals: ParseResults):
        """Draws a circle using radius. Chord_angle is ignored.
        Includes an automatic PD; restores the old pen state and returns
        to the center position afterwards."""
        vals_dict = vals.as_dict()
        r = vals_dict["Radius"]
        pendown = self.pendown
        # Circles are implicit subpolygons in polygon mode
        self._pm_command(1)

        self.pendown = False
        self.current_path.append(inkex.paths.move(-r, 0))
        self.pendown = True
        # sweep flag = 1 for counterclockwise rotation when y axis points upwards,
        # but this is honestly a best guess from an ambiguous specification
        self.current_path.append(inkex.paths.arc(r, r, 0, 1, 1, 2 * r, 0))
        self.current_path.append(inkex.paths.arc(r, r, 0, 1, 1, -2 * r, 0))
        self.pendown = False
        self.current_path.append(inkex.paths.move(r, 0))
        # Restore pendown
        self.pendown = pendown
        # and terminate subpolygon
        self._pm_command(1)

    def arc_command(self, vals: ParseResults):
        """Draws an absolute/ relative arc"""
        vals_dict = vals.as_dict()

        absolute = vals_dict["key"] == "AA"
        c_x, c_y = vals_dict["X"][0], vals_dict["Y"][0]
        sweep = vals_dict["sweep"] * math.pi / 180
        # Transform start point about center by sweep radius
        if not absolute:
            c_x, c_y = c_x + self.x, c_y + self.y
        dx, dy = (self.x - c_x), (self.y - c_y)
        endpoint = (
            dx * math.cos(sweep) - dy * math.sin(sweep) + c_x,
            dx * math.sin(sweep) + dy * math.cos(sweep) + c_y,
        )
        radius = math.sqrt(dx**2 + dy**2)
        # Determine SVG arc flags
        sweep_flag = 1 if sweep > 0 else 0
        large_arc = 1 if abs(sweep) > math.pi else 0
        res: inkex.paths.PathCommand = inkex.paths.Arc(
            radius, radius, 0, large_arc, sweep_flag, *endpoint
        )
        if not absolute:
            res = res.to_relative(inkex.Vector2d(self.x, self.y))
        self.current_path.append(res)

    def at_command(self, vals: ParseResults):
        """Draws an Absolute Three Point arc"""

        valsd = vals.as_dict()
        x, y, z = [
            i + 1j * j for i, j in zip([self.x] + valsd["X"], [self.y] + valsd["Y"])
        ]
        pel = inkex.PathElement.arc_from_3_points(x, y, z, "arc")
        path = pel.path
        path = path.to_absolute()
        if not list(path.end_points)[-1].is_close(z):
            path = path.reverse()
        assert list(path.end_points)[-1].is_close(z), list(path.end_points)

        for com in path[1:]:  # skip initial move command
            self.current_path.append(com)

    def bezier_command(self, vals: ParseResults):
        """Draws an absolute/relative bezier, possibly multiple"""
        absolute = vals["key"] == "BZ"
        vals = vals.as_dict()["B"]

        cmd = inkex.paths.Curve if absolute else inkex.paths.curve
        for group in vals:
            self.current_path.append(cmd(*group))

    def polyline_encoded(self, vals: ParseResults):
        """Parse Polyline Encoded"""
        data = vals["data"].encode("latin-1")
        absolute_mode = self.absolute_mode
        self.absolute_mode = False
        PolylineEncodedParser(self).polyline_encoded(data)

        self.absolute_mode = absolute_mode

    def pm_command(self, vals: ParseResults):
        """Entering / exiting polygon mode"""
        val = vals["value"]
        self._pm_command(int(val))

    def _pm_command(self, val: int):
        """Entering / exiting polygon mode"""
        if val == 0:
            self.polygon_mode = True
            self.finalize_path()
            # Link polygon buffer and current path.
            self.polygon_buffer = [self.current_path]
        if val > 0 and self.polygon_mode:
            # Move to the end point of the current polygon buffer
            if val == 2:
                self.polygon_mode = False

            if len(self.current_path) < 2:
                return

            self.initialize_current_path()
            if val == 1:
                self.polygon_buffer += [self.current_path]

    def ft_command(self, vals: ParseResults):
        """Fill type"""
        self.style_manager.ft_command(vals.as_dict())

    def _get_function(self, vals: ParseResults, obj):
        return getattr(obj, f"{vals[0][0].lower()}_command")

    def style_command(self, vals: ParseResults):
        """Handle all commands in the style groups, they are deferred to the
        StyleManager"""
        func = self._get_function(vals, self.style_manager)

        # For commands that take effect immediately, finalize the path. This is
        # managed by a decorator to keep all information in one place
        if hasattr(func, "clear_path"):
            self.finalize_path()

        func(vals[0].as_dict())

    def transform_command(self, vals: ParseResults):
        """Handle all commands that potentially create a new parent group"""
        self.finalize_path()
        func = self._get_function(vals, self.transform_manager)
        func(vals[0].as_dict())
        self.create_group()

    def edge_fill_polygon(self, vals: ParseResults):
        """Fill or edge polygon"""
        if self.polygon_mode:
            return  # command illegal in polygon mode
        pel = inkex.PathElement()
        pel.path = self.create_path(self.polygon_buffer, vals["key"] == "FP", True)
        self.current_parent.add(pel)
        if vals["key"] == "FP":
            self.style_manager.set_fill(pel)
            if vals["fillmode"] == "0":
                pel.style["fill-rule"] = "evenodd"
            else:
                pel.style["fill-rule"] = "nonzero"
        else:
            # Draw polygon edges
            pel.style["fill"] = "none"
            self.style_manager.set_stroke(
                pel, p1=self.transform_manager.p1, p2=self.transform_manager.p2
            )

    def rectangle_wedge(self, element, stroke, fill):
        """Common functionality of rectangles and wedges"""
        # Finish the current path
        self.finalize_path()
        self.current_parent.append(element)
        if stroke:
            element.style["fill"] = "none"
            self.style_manager.set_stroke(
                element, p1=self.transform_manager.p1, p2=self.transform_manager.p2
            )
        if fill:
            self.style_manager.set_fill(element)
        # This command does not change the current pen position and up/down state,
        # but it uses the polygon buffer internally. In case the buffer will be reused,
        # push the path to the polygon buffer
        pdata = inkex.Path(element.get_path()).to_absolute()
        # Omit closing Z since polygon buffer does this automatically
        self.polygon_buffer = [
            [  # type: ignore
                [i.command, i.command.letter != "M"] + list(i.first_point)
                for i in pdata.proxy_iterator()
                if i.command.letter.lower() != "z"
            ]
        ]

    def rectangle_command(self, vals: ParseResults):
        """Handle ER, EA, RA, RR"""
        command = vals["key"]
        x = vals["X"][0]
        y = vals["Y"][0]
        if command[1] == "R":
            x += self.x
            y += self.y
        rect = inkex.Rectangle.new(
            min(self.x, x), min(self.y, y), abs(self.x - x), abs(self.y - y)
        ).to_path_element()
        self.rectangle_wedge(rect, command[0] == "E", command[0] == "R")

    def wedge_command(self, vals: ParseResults):
        """Handle EW and WG commands"""
        command = vals["key"]
        radius = vals["radius"]
        start_angle = vals["start_angle"] + (0 if radius > 0 else 180)

        arc = inkex.PathElement.arc(
            [self.x, self.y],
            abs(vals["radius"]),
            arctype="slice",
            start=(start_angle % 360) * math.pi / 180,
            end=((start_angle + vals["sweep"]) % 360) * math.pi / 180,
        )

        self.rectangle_wedge(arc, command[0] == "E", command[0] == "W")


def clear_path(fun):
    """A decorator that tells the parent state machine that the path should be cleared
    before executing this method"""

    def wrapped(*args, **kwargs):
        return fun(*args, **kwargs)

    wrapped.clear_path = True  # type: ignore
    return wrapped


class StyleManager:
    """Abstraction layer to store the state from all commands affecting style (stroke
    and fill)
    Currently not implemented:
    - Fill types (patterns) except solid fill
    - fill anchor (AC)
    - raster fill (RF)
    - screened vectors (SV)
    - Symbol mode (SM)
    - carrying over residue of dasharray
    - adaptive line patterns (negative line indices)
    - triangle line cap
    - no-join (overlap) line join
    """

    def __init__(self, options) -> None:
        self.current_style = inkex.Style()
        self.fill_type = "solid"
        self.options = options
        self.width_units_relative: bool
        self.pen_width: float
        self.linetype_manager: LineTypeManager
        self.transparency: bool

    def initialize(self):
        """IN commands that affect the StyleManager"""
        self.width_units_relative = False
        self._reset_pen_width()
        self.sp_command({"pen": 0})

    def default_values(self):
        """DF commands that affect the StyleManager"""
        self.current_style["stroke-linecap"] = "butt"
        self.current_style["stroke-linejoin"] = "miter"
        self.current_style["stroke-miterlimit"] = 5
        self.fill_type = "solid"
        self.linetype_manager = LineTypeManager(None, True, 0.04)
        self.transparency = True

    @clear_path
    def sp_command(self, vals: Dict):
        """Select pen"""
        penmap = {
            0: "White",
            1: "Black",
            2: "Red",
            3: "Green",
            4: "Yellow",
            5: "Blue",
            6: "Magenta",
            7: "Cyan",
        }
        self.current_style["stroke"] = inkex.Color(penmap[vals.get("pen", 0)].lower())

    def ft_command(self, vals: Dict):
        """Fill type"""
        typ = int(vals.get("type", 1))
        if typ < 3:
            self.fill_type = "solid"

    @clear_path
    def la_command(self, vals: Dict):
        """Set line attributes (line ends, line joins)"""
        for k, value in zip(vals["kind"], vals["value"]):
            kind = int(k)
            # Triangle line ends / line joins are not supported
            if kind == 1:
                dct = {1: "butt", 2: "square", 3: "round", 4: "round"}
                self.current_style["stroke-linecap"] = dct[value]
            if kind == 2:
                #  6: no join is not supported; would require splitting the paths
                # could be done by inserting "m 0,0" after every command
                dct = {1: "miter", 2: "miter", 3: "round", 4: "round", 5: "bevel"}
                self.current_style["stroke-linejoin"] = dct.get(value, "miter")
            if kind == 3:
                self.current_style["stroke-miterlimit"] = value

    @clear_path
    def lt_command(self, vals: Dict):
        """Set line type"""
        self.linetype_manager.lt_command(vals)

    def _reset_pen_width(self):
        self.pen_width = 0.35 if not self.width_units_relative else 0.001

    @clear_path
    def pw_command(self, vals: Dict):
        """Set pen width"""
        if "width" not in vals:
            self._reset_pen_width()
        else:
            self.pen_width = vals["width"]

    def tr_command(self, vals: Dict):
        """Set transparency mode"""
        self.transparency = vals.get("transparency", "1") == "1"

    @clear_path
    def ul_command(self, vals: Dict):
        """Set user defined line type"""
        self.linetype_manager.ul_command(vals)

    def wu_command(self, vals: Dict):
        """Set pen width units"""
        self.width_units_relative = vals.get("units", "0") == "1"
        self._reset_pen_width()

    def p1_p2_mm(
        self,
        pt1: Union[Tuple[float, float], inkex.Vector2d],
        pt2: Union[Tuple[float, float], inkex.Vector2d],
    ):
        """Compute distance between p1 and p2 (provided in kwargs), and convert to mm"""
        return (
            math.sqrt((pt1[1] - pt2[1]) ** 2 + (pt1[0] - pt2[0]) ** 2)
            / 40
            * 1016
            / self.options.resolution
        )

    def set_stroke(
        self,
        element: inkex.PathElement,
        p1: Union[Tuple[float, float], inkex.Vector2d],
        p2: Union[Tuple[float, float], inkex.Vector2d],
    ):
        """Set stroke on an element based on the current stroke settings"""
        element.style["fill"] = None
        for i in ["dasharray", "linejoin", "miterlimit", "linecap"]:
            element.style[f"stroke-{i}"] = self.current_style(f"stroke-{i}")
        element.style["stroke"] = self.current_style("stroke")
        if self.transparency and self.current_style("stroke") == inkex.Color("white"):
            element.style["stroke-opacity"] = 0
        # Stroke width
        width = self.pen_width * (
            self.p1_p2_mm(p1, p2) if self.width_units_relative else 1
        )
        element.style["stroke-width"] = width
        # Apply stroke pattern
        self.linetype_manager.apply_to_path(element, self.p1_p2_mm(p1, p2))

    def set_fill(self, element):
        """Set stroke on an element based on the current fill settings.
        Currently, only solid fill is supported"""
        # Fill polygon
        if self.fill_type == "solid":
            element.style["fill"] = self.current_style("stroke")
            if self.transparency and self.current_style("stroke") == inkex.Color(
                "white"
            ):
                element.style["fill-opacity"] = 0


class LineTypeManager:
    """Wrapper for the information contained in the LT and UL command"""

    default_linetypes: List[List[Union[float, int]]] = [
        [0, 100],
        [50, 50],
        [70, 30],
        [80, 10, 0, 10],
        [70, 10, 10, 10],
        [50, 10, 10, 10, 10, 10],
        [70, 10, 0, 10, 0, 10],
        [50, 10, 0, 10, 10, 10, 0, 10],
    ]

    def __init__(
        self, index: Optional[int], unit_relative: bool, pattern_length: float
    ) -> None:
        # type, relative length, pattern length
        self.index = index
        self.unit_relative = unit_relative
        self.pattern_length = pattern_length

        self.linetypes = self.default_linetypes.copy()

    def lt_command(self, vals: Dict):
        """Set line type"""
        if len(vals) == 0:
            # no parameters, reset to solid lines
            self.index = None
        if "linetype" in vals:
            ltp = max(min(vals["linetype"], 8), 0)
            self.index = ltp
        if "mode" in vals:
            self.unit_relative = vals["mode"] == 0
        if "pattern_length" in vals:
            self.pattern_length = max(0, vals["pattern_length"])

    def ul_command(self, vals: Dict):
        """Set user defined line type"""
        if "index" not in vals:
            # Reset all line types:
            self.linetypes = self.default_linetypes.copy()
            return
        i = int(vals["index"])
        if i < 1 or i > 8:
            return
        if "gap" not in vals:
            # Reset current line type
            self.linetypes[i - 1] = self.default_linetypes[i - 1]
            return
        gaps: List[float] = vals["gap"]
        if any(gap < 0 for gap in gaps) or sum(gaps) <= 0:
            return  # invalid
        self.linetypes[i - 1] = [gap / sum(gaps) * 100 for gap in gaps]

    def apply_to_path(self, element: inkex.PathElement, p1p2_dist):
        """Apply the line style to a path element."""
        if self.index is not None:  # None -> solid line
            pattern_length = self.pattern_length * (
                p1p2_dist / 100 if self.unit_relative else 1
            )
            if self.index > 0:
                element.style["stroke-dasharray"] = " ".join(
                    [
                        str(dashdot / 100.0 * pattern_length)
                        for dashdot in self.linetypes[self.index - 1]
                    ]
                )
            elif self.index == 0:
                # Only keep the end points of the path.
                new_path = inkex.Path()
                path_data = list(element.path.proxy_iterator())
                for i, item in enumerate(path_data):
                    if item.command.letter.lower() == "m":
                        if path_data[i + 1].command.letter.lower() != "m":
                            new_path.append(inkex.paths.Move(*(item.end_point)))
                            new_path.append(inkex.paths.line(0.0001, 0))
                    else:
                        # Workaround https://gitlab.com/inkscape/inkscape/-/issues/894
                        new_path.append(inkex.paths.Move(*(item.end_point)))
                        new_path.append(inkex.paths.line(0.0001, 0))
                element.path = new_path
                # Negative line types would have to be implemented in a similar way


class TransformManager:
    """Abstraction layer to handle all commands that affect transform of the current
    layer, in particular IP, IR, SC, IW, RO"""

    def __init__(self, options) -> None:
        self.scale_manager = ScaleManager()
        self.clip_manager = ClipManager()
        self.p1: inkex.Vector2d
        self.p2: inkex.Vector2d
        self.ro_angle: int = 0
        self.options = options
        self.ir_command = self.ip_command  # Alias
        self.initialize()

    def initialize(self):
        """IN commands that affect the TransformManager"""
        self.default_values()
        self.p1 = inkex.Vector2d((0, 0))
        self.p2 = inkex.Vector2d(
            (
                self.options.width * self.options.resolution,
                self.options.height * self.options.resolution,
            )
        )
        self.ro_angle = 0

    def default_values(self):
        """DF commands that affect the TransformManager"""
        self.scale_manager = ScaleManager(None)

    def get_transform_and_clip(
        self, root_transform: inkex.Transform
    ) -> Tuple[inkex.Transform, Optional[inkex.ClipPath]]:
        """Return the current effective transform and clip"""
        transform = self.get_rotation_transform() @ self.scale_manager.get_transform(
            self.p1, self.p2
        )
        clip = self.clip_manager.get_current_clip(
            root_transform,
            transform,
            scaled=self.scale_manager.enabled,
            bake=self.options.bake_transforms,
        )
        return transform, clip

    def ip_command(self, vals: Dict):
        """Set the scaling points p1 and p2"""
        key = vals["key"]
        x = [float(i) for i in vals["X"]]
        y = [float(i) for i in vals["Y"]]

        if key == "IR":
            x = [i * self.options.width * self.options.resolution / 100 for i in x]
            y = [i * self.options.height * self.options.resolution / 100 for i in y]
        if len(x) == 1:
            # P2 tracks P1
            x = [x[0], self.p2[0] - self.p1[0] + x[0]]
            y = [y[0], self.p2[1] - self.p1[1] + y[0]]
        for i in x, y:
            if i[0] == i[1]:  # avoid scaling into infinity
                i[1] += 1

        self.p1 = inkex.Vector2d(x[0], y[0])
        self.p2 = inkex.Vector2d(x[1], y[1])

    def sc_command(self, vals: Dict):
        """Scale command"""
        self.clip_manager.pin_to_plu()
        self.scale_manager = ScaleManager(vals)

    def get_rotation_transform(self) -> inkex.Transform:
        """Compute the effective transform caused by the RO command.
        Does not currently handle a following IP/IR without arguments."""
        rotation = inkex.Transform().add_rotate(self.ro_angle, self.p1)
        translation = inkex.Transform()
        if self.ro_angle == 90:
            # Shift P1 into the bottom right corner
            translation.add_translate(self.p2[0] - self.p1[0], 0)
        elif self.ro_angle == 180:
            # Shift P1 into the top right corner
            translation.add_translate(self.p2[0] - self.p1[0], self.p2[1] - self.p1[1])
        elif self.ro_angle == 270:
            # Shift P1 into the top left corner
            translation.add_translate(0, self.p2[1] - self.p1[1])
        return translation @ rotation

    def ro_command(self, vals: Dict):
        """Handle canvas rotation"""
        self.ro_angle = int(vals.get("angle", 0))

    def iw_command(self, vals: Dict):
        """Handle Soft Clipping"""
        self.clip_manager = ClipManager(vals.get("X", None), vals.get("Y", None))


class ScaleManager:
    """Helper class for the SC command"""

    def __init__(self, vals: Optional[Dict] = None):
        if vals is None:
            vals = {}
        self.vals = vals

    @property
    def enabled(self):
        """Determine if scaling is switched on"""
        return "Xmin" in self.vals

    def get_transform(self, p1, p2) -> inkex.Transform:
        """Return the transform from plotter units to user units, using the scaling
        points p1 and p2"""
        if not self.enabled:
            return inkex.Transform()
        xmin = self.vals["Xmin"]
        xmax = self.vals["Xmax"]
        ymin = self.vals["Ymin"]
        ymax = self.vals["Ymax"]
        # default not clear from spec? prob anisotropic
        type_id = self.vals.get("type", "0")

        if type_id != "2":
            if type_id == "0":
                pmin = [xmin, ymin]
                pmax = [xmax, ymax]
            elif type_id == "1":
                scale_y = abs(p2[1] - p1[1]) / abs(ymax - ymin)
                scale_x = abs(p2[0] - p1[0]) / abs(xmax - xmin)
                scale = min(scale_x, scale_y)
                if scale == scale_y:
                    # Space left / right
                    space = abs(p2[0] - p1[0]) - scale * abs(xmax - xmin)
                    left = (self.vals.get("left", 50) / 100 * space) / scale
                    right = ((100 - self.vals.get("left", 50)) / 100 * space) / scale
                    pmin = [xmin - left, ymin]
                    pmax = [xmax + right, ymax]
                if scale == scale_x:
                    # Space left / right
                    space = abs(p2[1] - p1[1]) - scale * abs(ymax - ymin)
                    bottom = (self.vals.get("bottom", 50) / 100 * space) / scale
                    top = ((100 - self.vals.get("bottom", 50)) / 100 * space) / scale
                    pmin = [xmin, ymin - bottom]
                    pmax = [xmax, ymax + top]
            # We need to map the points (xmin, ymin), (xmax, ymax))
            # in the layer coordinate system to  (P1), (P2) in plotter units,
            scale = [(p1[i] - p2[i]) / (pmin[i] - pmax[i]) for i in range(2)]
            offset = [p1[i] - scale[i] * pmin[i] for i in range(2)]
        elif type_id == "2":
            pmin = [xmin, ymin]
            scale = [xmax, ymax]
            offset = [p1[0] - scale[0] * xmin, p1[1] - scale[1] * ymin]
        return inkex.Transform(scale=tuple(scale), translate=tuple(offset))


class ClipManager:
    """Helper class for the IW command"""

    def __init__(
        self,
        x: Optional[Tuple[float, float]] = None,
        y: Optional[Tuple[float, float]] = None,
    ) -> None:
        self.x = x
        self.y = y
        self.storedx: Optional[Tuple[float, float]] = None
        self.storedy: Optional[Tuple[float, float]] = None
        self.pinned = False

    def get_current_clip(
        self,
        svg_to_plu: inkex.Transform,
        plu_to_uu: inkex.Transform,
        scaled=False,
        bake=False,
    ) -> Optional[inkex.ClipPath]:
        """Get the current clip path based on current transforms and scaling info"""
        if self.x is None or self.y is None:
            return None
        if not (self.x[1] >= self.x[0] and self.y[1] >= self.y[0]):
            inkex.errormsg("Bad clipping specification, will be ignored")
            return None
        if self.pinned:
            transform = svg_to_plu
        else:
            transform = svg_to_plu @ plu_to_uu
        if not bake:
            transform = -(svg_to_plu @ plu_to_uu) @ transform

        rect = inkex.Rectangle.new(
            self.x[0], self.y[0], self.x[1] - self.x[0], self.y[1] - self.y[0]
        )
        aspath = inkex.Path(rect.get_path())

        if scaled:
            # Store positions of the rectangle in plotter units
            bbox = aspath.transform(plu_to_uu).bounding_box()
            self.storedx = (float(bbox.left), float(bbox.right))
            self.storedy = (float(bbox.top), float(bbox.bottom))

        result = inkex.ClipPath()
        result.add(inkex.PathElement.new(path=aspath.transform(transform)))
        return result

    def pin_to_plu(self):
        """Pin position of the clip in plotter units"""
        self.x = self.storedx
        self.y = self.storedy
        self.pinned = True


class PolylineEncodedParser:
    """Helper class to parse the PE command"""

    def __init__(self, parent: HPGLStateMachine):
        self.parent = parent

    @staticmethod
    def _decode_value(value, fractional_bits: Optional[int] = None, base32_mode=False):
        """Helper function to decode a HPGL-encoded number"""
        # First remove all irrelevant values
        value = "".join(
            chr(i) for i in value if not (i < 63 or 128 <= i <= 190 or i == 255)
        ).encode("latin-1")
        result = 0
        for i, val in enumerate(value):
            if base32_mode and val > 127:
                val -= 128
            offset = 63
            if i == len(value) - 1:
                if base32_mode:
                    offset = 95
                else:
                    offset = 191

            result += (val - offset) << (5 if base32_mode else 6) * i
        # Only perform these steps if fractional data (i.e. coordinates)
        # are to be encoded
        if fractional_bits is not None:
            # Set sign
            result = result // 2 * (-1 if result % 2 == 1 else 1)
            # Fractional bits
            result = result / (2**fractional_bits)
        return result

    @staticmethod
    def _get_next_value(
        data, start_with, fractional_bits: Optional[int] = None, base32_mode=False
    ):
        """Helper function to find and then decode a HPGL encoded number"""
        i = start_with
        while i < len(data):
            if chr(data[i]) in ":><=7":  # can happen if there is no data
                # (specification is contradictory if this is allowed or not)
                i -= 1
                break
            if base32_mode:
                if data[i] % 128 >= 95:
                    break
            else:
                if data[i] >= 191:
                    break
            i += 1
        return i + 1, PolylineEncodedParser._decode_value(
            data[start_with : i + 1], fractional_bits, base32_mode
        )

    def polyline_encoded(self, data):
        """Parse a binary-encoded polyline"""
        index = 0
        base32_mode = False
        fractional_bits = 0

        while index < len(data):
            flag = chr(data[index])
            if base32_mode:
                flag = chr(ord(flag) % 128)  # remove highest bit
            if flag == "7":
                base32_mode = True
                index += 1
            elif flag in ">:":
                index, result = PolylineEncodedParser._get_next_value(
                    data, index + 1, None, base32_mode
                )
                if flag == ":":
                    # Select pen
                    self.parent.style_manager.sp_command({"pen": result})
                else:
                    # Select the number of fractional bits
                    fractional_bits = result
            else:
                start_index = index + (1 if flag in "<=" else 0)
                index, first = PolylineEncodedParser._get_next_value(
                    data, start_index, fractional_bits, base32_mode
                )
                index, second = PolylineEncodedParser._get_next_value(
                    data, index, fractional_bits, base32_mode
                )
                cmd = "PU" if flag == "<" else ("PA" if flag == "=" else "PR")
                # pylint: disable=protected-access
                if start_index != index:
                    self.parent._line_command(cmd, [first], [second])
                    # If coordinates were found, send a pen-down command if necessary
                    if not self.parent.pendown:
                        self.parent._line_command("PD")
                else:
                    self.parent._line_command(cmd)



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2005 Pim Snel, pim@lingewoud.com
# Copyright (C) 2008 Aaron Spike, aaron@ekips.org
# Copyright (C) 2011 Nicolas Dufour, nicoduf@yahoo.fr
#
#    * Fix for a bug related to special characters in the path (LP #456248).
#    * Fix for Windows support (LP #391307 ).
#    * Font list and image directory features.
#
# this is  the first Python script  ever created
# its based on embedimage.py
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# TODOs
# - fix bug: not saving existing .zip after a Collect for Output is run
#     this bug occurs because after running an effect extension the inkscape:output_extension is reset to svg.inkscape
#     the file name is still xxx.zip. after saving again the file xxx.zip is written with a plain .svg which
#     looks like a corrupt zip
# - maybe add better extension
# - consider switching to lzma in order to allow cross platform compression with no encoding problem...
#
"""
An extension which collects all images to the documents directory and
creates a zip archive containing all images and the document
"""

import os
import tempfile
import zipfile

from typing import List, Tuple
from urllib.parse import urlparse
from urllib.request import url2pathname

import inkex
from inkex import TextElement, Tspan, FlowRoot, FlowPara, FlowSpan
from inkex.localization import inkex_gettext as _


class CompressedMedia(inkex.OutputExtension):
    """Output a compressed file"""

    def __init__(self):
        super().__init__()
        self.path_dict = {}

    def add_arguments(self, pars):
        pars.add_argument("--image_dir", help="Image directory", default="images")
        pars.add_argument("--font_list", type=inkex.Boolean, help="Add font list")

    def process_path(self, path: str) -> Tuple[str, bool]:
        """Processes an absolute path and returns
        (unique filename, exists)."""

        if path in self.path_dict:
            return os.path.split(path)[1], True

        index = 0
        pth, ext = os.path.splitext(path)
        _, filename = os.path.split(pth)
        trypath = filename + ext
        while True:
            if trypath in list(self.path_dict.values()):
                index += 1
                trypath = f"{filename}{index}{ext}"
            else:
                self.path_dict[path] = trypath
                return trypath, False

    def collect_images(self, docname, z: zipfile.ZipFile):
        """
        Collects all images in the document
        and copy them to the temporary directory.
        """
        imgdir = self.options.image_dir

        for node in self.svg.xpath("//svg:image"):
            xlink = node.get("xlink:href")
            if xlink[:4] != "data":
                url = urlparse(xlink)
                href = url2pathname(url.path)

                image_path = self.absolute_href(href or "")

                # Backup directory where we can find the image
                if not os.path.isfile(image_path):
                    image_path = node.get("sodipodi:absref", image_path)

                if not os.path.isfile(image_path):
                    inkex.errormsg(
                        _('File not found "{}". Unable to embed image.').format(
                            image_path
                        )
                    )
                    continue
                else:
                    zippath, exists = self.process_path(image_path)
                    if not exists:
                        z.write(image_path, os.path.join(imgdir, zippath))

                    node.set("xlink:href", f"{imgdir}/{zippath}")

    def collect_svg(self, docstripped, z: zipfile.ZipFile):
        """
        Copy SVG document to the temporary directory
        and add it to the temporary compressed file
        """
        dst_file = os.path.join(self.tmp_dir, docstripped)
        with open(dst_file, "wb") as stream:
            self.document.write(stream)
        z.write(dst_file, docstripped + ".svg")

    def is_text(self, node):
        """
        Returns true if the tag in question is an element that
        can hold text.
        """
        return isinstance(node, (TextElement, Tspan, FlowRoot, FlowPara, FlowSpan))

    def get_fonts(self, node):
        """
        Given a node, returns a list containing all the fonts that
        the node is using.
        """
        fonts = []
        s = ""
        if "style" in node.attrib:
            s = dict(inkex.Style.parse_str(node.attrib["style"]))
        if not s:
            return fonts

        if "font-family" in s:
            if "font-weight" in s:
                fonts.append(s["font-family"] + " " + s["font-weight"])
            else:
                fonts.append(s["font-family"])
        elif "-inkscape-font-specification" in s:
            fonts.append(s["-inkscape-font-specification"])
        return fonts

    def list_fonts(self, z: zipfile.ZipFile):
        """
        Walks through nodes, building a list of all fonts found, then
        reports to the user with that list.
        Based on Craig Marshall's replace_font.py
        """
        nodes: List[inkex.BaseElement] = []
        items = self.svg.iterdescendants()
        nodes.extend(filter(self.is_text, items))
        fonts_found = []
        for node in nodes:
            for f in self.get_fonts(node):
                if not f in fonts_found:
                    fonts_found.append(f)
        findings = sorted(fonts_found)
        # Write list to the temporary compressed file
        filename = "fontlist.txt"
        dst_file = os.path.join(self.tmp_dir, filename)
        with open(dst_file, "w") as stream:
            if len(findings) == 0:
                stream.write(_("Didn't find any fonts in this document/selection."))
            else:
                if len(findings) == 1:
                    stream.write(_("Found the following font only: %s") % findings[0])
                else:
                    stream.write(
                        _("Found the following fonts:\n%s") % "\n".join(findings)
                    )
        z.write(dst_file, filename)

    def save(self, stream):
        docname = self.svg.get("sodipodi:docname")

        if docname is None:
            docname = self.options.input_file

        # TODO: replace whatever extension
        docstripped = os.path.basename(docname.replace(".zip", ""))
        docstripped = docstripped.replace(".svg", "")
        docstripped = docstripped.replace(".svgz", "")

        # Create os temp dir
        self.tmp_dir = tempfile.mkdtemp()

        # Create destination zip in same directory as the document
        with zipfile.ZipFile(stream, "w") as z:
            self.collect_images(docname, z)
            self.collect_svg(docstripped, z)
            if self.options.font_list:
                self.list_fonts(z)


if __name__ == "__main__":
    CompressedMedia().run()



#!/usr/bin/env python3
#
# Copyright (C) 2009 Aurelio A. Heckert, aurium (a) gmail dot com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import inkex
from inkex.localization import inkex_gettext as _

import inkwebeffect


class TransmitAttribute(inkwebeffect.InkWebEffect):
    def add_arguments(self, pars):
        pars.add_argument("--tab")
        pars.add_argument("--att", default="fill", help="Attribute to transmitted.")
        pars.add_argument("--when", default="onclick", help="When it must to transmit?")
        pars.add_argument("--from-and-to", dest="from_and_to", default="g-to-one")
        pars.add_argument(
            "--compatibility",
            default="append",
            help="Compatibility with previews code to this event.",
        )

    def effect(self):
        self.ensureInkWebSupport()

        if len(self.options.ids) < 2:
            raise inkex.AbortExtension(_("You must select at least two elements."))

        # All set the last else The first set all
        split = -1 if self.options.from_and_to == "g-to-one" else 1
        el_from = list(self.svg.selection)[:split]
        id_to = list(self.svg.selection.ids)[split:]

        ev_code = "InkWeb.transmitAtt({{from:this, to:['{}'], att:'{}'}})".format(
            "','".join(id_to), self.options.att
        )
        for elem in el_from:
            prev_ev_code = elem.get(self.options.when, "")
            if self.options.compatibility == "append":
                el_ev_code = prev_ev_code + ";\n" + ev_code
            if self.options.compatibility == "prepend":
                el_ev_code = ev_code + ";\n" + prev_ev_code
            if self.options.compatibility == "replace":
                el_ev_code = ev_code
            elem.set(self.options.when, el_ev_code)


if __name__ == "__main__":
    TransmitAttribute().run()



# coding=utf-8
#
# Copyright (C) 2021 Jonathan Neuhauser, jonathan.neuhauser@outlook.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
Property management and parsing, CSS cascading, default value storage

.. versionadded:: 1.2

.. data:: all_properties

    A list of all properties, their parser class, and additional information
    such as whether they are inheritable or can be given as presentation attributes
"""

from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass

from typing import (
    Dict,
    List,
    Optional,
    TYPE_CHECKING,
    cast,
)
import tinycss2
import tinycss2.ast

from .interfaces.IElement import IBaseElement

from .units import convert_unit
from .utils import FragmentError

from .colors import Color

if TYPE_CHECKING:
    from .elements import BaseElement

TokenList = List[tinycss2.ast.Node]


def _make_number_token_dimless(token):
    if isinstance(token, (tinycss2.ast.NumberToken, tinycss2.ast.PercentageToken)):
        return token.value
    if isinstance(token, tinycss2.ast.DimensionToken):
        return convert_unit(token.serialize(), "px")
    raise ValueError("Not a number")


def _get_tokens_from_value(value: str) -> TokenList:
    return tinycss2.parse_one_declaration(f"a:{value}").value


def _is_ws(token: tinycss2.ast.Node) -> bool:
    return isinstance(token, tinycss2.ast.WhitespaceToken)


def _strip_whitespace_nodes(value: TokenList) -> TokenList:
    try:
        while _is_ws(value[0]):
            del value[0]
        while _is_ws(value[-1]):
            del value[-1]
        return value
    except IndexError:
        return []


def _is_inherit(value: TokenList | None) -> bool:
    return (
        value is not None
        and len(value) == 1
        and isinstance(value[0], tinycss2.ast.IdentToken)
        and value[0].value == "inherit"
    )


class _StyleConverter:
    """Converter between str and computed value of a style, with context element"""

    # pylint: disable=unused-argument
    def convert(
        self, value: TokenList, element: Optional[BaseElement] = None
    ) -> object:
        """Get parsed property value with resolved urls, color, etc.

        Args:
            element (BaseElement): the SVG element to which this style is applied to
                currently used for resolving gradients / masks, could be used for
                computing percentage attributes or calc() attributes [optional]

        Returns:
            object: parsed property value
        """
        return tinycss2.serialize(value)

    def raise_invalid_value(
        self, value: TokenList, element: Optional[BaseElement] = None
    ) -> None:
        """Checks if the value str is valid in the context of element.

        Args:
            value (str): attribute value
            element (Optional[BaseElement], optional): Context element. Defaults to
                None.

        Raises:
            various exceptions if the property has a bad value
        """
        self.convert(value, element)

    def convert_back(
        self, value: object, element: Optional[BaseElement] = None
    ) -> TokenList:
        """Converts value back to string in the context of element.

        Args:
            value (object): parsed value of attribute
            element (_type_, optional): Context element. Defaults to None.

        Returns:
            str: _description_
        """
        return _get_tokens_from_value(str(value))


class _AlphaValueConverter(_StyleConverter):
    """Stores an alpha value (such as opacity), which may be specified as
    as percentage or absolute value.

    Reference: https://www.w3.org/TR/css-color/#typedef-alpha-value"""

    def convert(
        self, value: TokenList, element: Optional[BaseElement] = None
    ) -> object:
        if isinstance(value[0], tinycss2.ast.NumberToken):
            parsed_value = value[0].value
        elif isinstance(value[0], tinycss2.ast.PercentageToken):
            parsed_value = value[0].value / 100
        else:
            raise ValueError()
        return min(max(parsed_value, 0), 1)

    def convert_back(
        self, value: object, element: Optional[BaseElement] = None
    ) -> TokenList:
        if isinstance(value, (float, int)):
            value = min(max(value, 0), 1)
            return [tinycss2.ast.NumberToken(0, 0, value, value, f"{value}")]
        raise ValueError("Value must be number")


class _ColorValueConverter(_StyleConverter):
    """Converts a color value

    Reference: https://drafts.csswg.org/css-color-3/#valuea-def-color"""

    def convert(
        self, value: TokenList, element: Optional[BaseElement] = None
    ) -> object:
        # Process this as string
        vstr = _StyleConverter.convert(self, value, element)
        if vstr == "currentColor":
            if element is not None:
                return element.get_computed_style("color")
            return None
        return Color(vstr)


class _URLNoneValueConverter(_StyleConverter):
    """Converts a value that is either none or an url, such as markers or masks.

    Reference: https://www.w3.org/TR/SVG2/painting.html#VertexMarkerProperties"""

    def convert(
        self, value: TokenList, element: Optional[BaseElement] = None
    ) -> object:
        if len(value) == 0:
            return None
        value = value[0]
        if isinstance(value, tinycss2.ast.URLToken):
            if element is not None and self.element_has_root(element):
                return element.root.getElementById(value.value)
            return value.serialize()
        if isinstance(value, tinycss2.ast.IdentToken):
            return None

        raise ValueError("Invalid property value")

    def convert_back(
        self, value: object, element: Optional[BaseElement] = None
    ) -> TokenList:
        if isinstance(value, IBaseElement):
            if element is not None:
                value = _URLNoneValueConverter._insert_if_necessary(element, value)
            return [tinycss2.ast.URLToken(0, 0, value.get_id(), value.get_id(as_url=2))]
        return [tinycss2.ast.IdentToken(0, 0, "none")]

    @staticmethod
    def element_has_root(element) -> bool:
        "Checks if an element has a root"
        try:
            _ = element.root
            return True
        except FragmentError:
            return False

    @staticmethod
    def _insert_if_necessary(element: BaseElement, value: BaseElement) -> BaseElement:
        """Ensures that the return (value or a deep copy of it) is inside the same
        document as element.

        if element is unrooted, don't do anything (just return value)
        If element is attached to the same document as value, return value.
        If value is not attached to a document, attach it to the defs of element's
        document and return value.
        If value is already attached to another document, create a copy and return the
        copy.

        .. versionadded:: 1.4"""
        # Check if the element is rooted and has the same root as self.element.
        try:
            _ = element.root
            try:
                if value.root != element.root:
                    # Create a copy and attach it to the tree.
                    copy = value.copy()
                    element.root.defs.append(copy)
                    return copy
            except FragmentError:
                element.root.defs.append(value)
                return value

        except FragmentError:
            pass
        return value


class _PaintValueConverter(_ColorValueConverter, _URLNoneValueConverter):
    """Stores a paint value (such as fill and stroke), which may be specified
    as color, or url.

    Reference: https://www.w3.org/TR/SVG2/painting.html#SpecifyingPaint"""

    def convert(
        self, value: TokenList, element: Optional[BaseElement] = None
    ) -> object:
        v0 = value[0]
        if isinstance(v0, tinycss2.ast.IdentToken):
            if v0.value == "none":
                return None
            if v0.value == "currentColor":
                return _ColorValueConverter.convert(self, value, element)
            if v0.value in ["context-fill", "context-stroke"]:
                return v0.value
            else:
                return Color(v0.value)

        if isinstance(v0, tinycss2.ast.HashToken):
            return Color("#" + v0.value)
        if isinstance(v0, tinycss2.ast.URLToken):
            if element is not None and self.element_has_root(element):
                paint_server = element.root.getElementById(v0.value[1:])
            else:
                return None
            if paint_server is not None:
                return paint_server
            for i in value[1:]:
                if isinstance(i, tinycss2.ast.IdentToken):
                    return Color(i.value)
            raise ValueError("Paint server not found")
        if isinstance(v0, tinycss2.ast.FunctionBlock) and v0.name in [
            "rgb",
            "rgba",
            "hsl",
            "hsla",
        ]:
            arguments = [str(argument.value) for argument in v0.arguments]
            return Color(f"{v0.name}({''.join(arguments)})")
        raise ValueError("Unknown color specification")

    def convert_back(
        self, value: object, element: Optional[BaseElement] = None
    ) -> TokenList:
        if value is None:
            return [tinycss2.ast.IdentToken(0, 0, "none")]
        if isinstance(value, IBaseElement):
            return _URLNoneValueConverter.convert_back(self, value, element=element)
        return _ColorValueConverter.convert_back(self, value, element=element)


class _EnumValueConverter(_StyleConverter):
    """Stores a value that can only have a finite set of options"""

    def __init__(self, options):
        self.options = options

    def raise_invalid_value(
        self, value: TokenList, element: BaseElement | None = None
    ) -> None:
        if tinycss2.serialize(value) not in self.options:
            raise ValueError(
                f"Value '{tinycss2.serialize(value)}' is invalid for the property"
            )


class _ShorthandValueConverter(_StyleConverter):
    """Stores a value that sets other values (e.g. the font shorthand)"""

    def __init__(self, keys) -> None:
        super().__init__()

        self.keys = keys

    @abstractmethod
    def get_shorthand_changes(self, value) -> Dict[str, str]:
        """calculates the value of affected attributes for this shorthand

        Returns:
            Dict[str, str]: a dictionary containing the new values of the
            affected attributes
        """


class _FontValueShorthandConverter(_ShorthandValueConverter):
    """Logic for the shorthand font property"""

    def __init__(self) -> None:
        super().__init__(
            [
                "font-style",
                "font-variant",
                "font-weight",
                "font-stretch",
                "font-size",
                "line-height",
                "font-family",
            ]
        )
        self.options = {
            key: all_properties[key].converter.options  # type: ignore
            for key in self.keys
            if isinstance(all_properties[key].converter, _EnumValueConverter)
        }
        # Font stretch can be specified in percent, but for the
        # shorthand, only a keyword value is allowed
        self.options["font-stretch"] = (
            "normal",
            "ultra-condensed",
            "extra-condensed",
            "condensed",
            "semi-condensed",
            "semi-expanded",
            "expanded",
            "extra-expanded",
            "ultra-expanded",
        )

    def get_shorthand_changes(self, value):
        result = {key: all_properties[key].default_value for key in self.keys}

        if len(value) == 0:
            return result  # shorthand not set, nothing to do
        i = 0
        while i < len(value):
            cur = value[i]
            matched = False
            if isinstance(cur, tinycss2.ast.IdentToken):
                matched = True
                if cur.value in self.options["font-style"]:
                    result["font-style"] = [cur]
                elif cur.value in self.options["font-variant"]:
                    result["font-variant"] = [cur]
                elif cur.value in self.options["font-weight"]:
                    result["font-weight"] = [cur]
                elif cur.value in self.options["font-stretch"]:
                    result["font-stretch"] = [cur]
                else:
                    matched = False
            if not matched and not isinstance(cur, tinycss2.ast.WhitespaceToken):
                result["font-size"] = [cur]
                if (
                    len(value) > i + 1
                    and isinstance(value[i + 1], tinycss2.ast.LiteralToken)
                    and value[i + 1].value == "/"
                ):
                    result["line-height"] = [value[i + 2]]
                    i += 2
                if len(value[i + 1 :]) > 0:
                    result["font-family"] = _strip_whitespace_nodes(value[i + 1 :])
                break
            i += 1
        return result


class _TextDecorationValueConverter(_ShorthandValueConverter):
    """Logic for the shorthand font property

    .. versionadded:: 1.3"""

    def __init__(self):
        super().__init__(
            ["text-decoration-style", "text-decoration-color", "text-decoration-line"]
        )
        self.options = {
            "text-decoration-" + key: all_properties[
                "text-decoration-" + key
            ].converter.options
            for key in ("line", "style", "color")
            if isinstance(
                all_properties["text-decoration-" + key].converter, _EnumValueConverter
            )
        }

    def get_shorthand_changes(self, value):
        result = {
            "text-decoration-style": all_properties[
                "text-decoration-style"
            ].default_value,
            "text-decoration-color": _get_tokens_from_value("currentcolor"),
            "text-decoration-line": [],
        }

        for token, cur in list((i, i.serialize()) for i in value):
            if cur in ["underline", "overline", "line-through", "blink"]:
                result["text-decoration-line"].extend(
                    [token, tinycss2.ast.WhitespaceToken(0, 0, value=" ")]
                )
            elif cur in self.options["text-decoration-style"]:
                result["text-decoration-style"] = [token]
            elif cur.strip():
                result["text-decoration-color"] = [token]

        if len(result["text-decoration-line"]) == 0:
            result["text-decoration-line"] = all_properties[
                "text-decoration-line"
            ].default_value
        else:
            result["text-decoration-line"] = result["text-decoration-line"][:-1]

        return result


class _MarkerShorthandValueConverter(_ShorthandValueConverter, _URLNoneValueConverter):
    """Logic for the marker shorthand property"""

    def __init__(self) -> None:
        super().__init__(["marker-start", "marker-end", "marker-mid"])

    def get_shorthand_changes(self, value):
        if value == "":
            return {}  # shorthand not set, nothing to do
        return {k: value for k in self.keys}

    def convert_back(
        self, value: object, element: Optional[BaseElement] = None
    ) -> TokenList:
        """Convert value back to a tokenlist"""
        return _URLNoneValueConverter.convert_back(self, value, element)


class _FontSizeValueConverter(_StyleConverter):
    """Logic for the font-size property"""

    def convert(
        self, value: TokenList, element: Optional[BaseElement] = None
    ) -> object:
        v0 = value[0].serialize()
        if element is None:
            return v0  # no additional logic in this case
        if isinstance(
            value[0],
            (
                tinycss2.ast.NumberToken,
                tinycss2.ast.PercentageToken,
                tinycss2.ast.DimensionToken,
            ),
        ):
            return element.to_dimensionless(v0)
        # unable to parse font size, e.g. font-size:normal
        return v0


class _StrokeDasharrayValueConverter(_StyleConverter):
    """Logic for the stroke-dasharray property"""

    def convert(
        self, value: TokenList, element: Optional[BaseElement] = None
    ) -> object:
        dashes = []
        i = 0
        for i, el in enumerate(value):
            if (
                i == 0
                and isinstance(el, tinycss2.ast.IdentToken)
                and el.value == "none"
            ):
                return []
            # whitespace or comma separated list
            if isinstance(el, (tinycss2.ast.WhitespaceToken)) or (
                isinstance(el, tinycss2.ast.LiteralToken) and el.value == ","
            ):
                continue
            if isinstance(el, (tinycss2.ast.DimensionToken, tinycss2.ast.NumberToken)):
                dashes.append(_make_number_token_dimless(el))
            else:
                return []
        if any(i < 0 for i in dashes):
            # one negative value makes the dasharray invalid
            return []
        if len(dashes) % 2 == 1:
            dashes = 2 * dashes
        return dashes

    def convert_back(
        self, value: object, element: Optional[BaseElement] = None
    ) -> TokenList:
        if value is None or not value:
            value = "none"
        if isinstance(value, list):
            if len(value) == 0:
                value = "none"
            else:
                value = " ".join(map(str, value))
        return _get_tokens_from_value(str(value))


class _FilterListConverter(_URLNoneValueConverter):
    """Stores a list of Filters

    .. versionadded:: 1.4"""

    def convert(
        self, value: TokenList, element: Optional[BaseElement] = None
    ) -> object:
        if element is None or value == "none":
            return []

        try:
            result = [
                element.root.getElementById(item.value)
                for item in value
                if isinstance(item, tinycss2.ast.URLToken)
            ]
            return [i for i in result if i is not None]
        except FragmentError:
            return [
                item.serialize()
                for item in value
                if isinstance(item, tinycss2.ast.URLToken)
            ]
        except ValueError:  # broken link
            return []

    def convert_back(
        self, value: object, element: Optional[BaseElement] = None
    ) -> TokenList:
        if isinstance(value, IBaseElement) or not isinstance(value, (list, tuple)):
            value = [value]
        if all((isinstance(i, str) for i in value)):
            return _get_tokens_from_value(" ".join(value))  # type: ignore
        assert element is not None
        value = cast("List[BaseElement]", value)
        try:
            _ = element.root
            for i in value:
                if i is None:
                    continue
                # insert the element
                inserted = self._insert_if_necessary(element, cast("BaseElement", i))
                # if a copy was created, replace the original in the list with the copy
                if inserted is not i:
                    for index, item in enumerate(value):
                        if item is i:
                            value[index] = inserted
        except FragmentError:
            # Element is unrooted, we skip this step.
            pass
        return _get_tokens_from_value(
            " ".join(f"url(#{i.get_id()})" for i in value if i is not None)
        )


@dataclass
class PropertyDescription:
    """Describes a CSS / presentation attribute"""

    name: str
    converter: _StyleConverter
    default_value: TokenList
    presentation: bool
    inherited: bool

    def __init__(
        self,
        name: str,
        converter: _StyleConverter,
        default_value: str,
        presentation: bool = False,
        inherited: bool = False,
    ):
        self.presentation = presentation
        self.inherited = inherited
        self.default_value = _get_tokens_from_value(default_value)
        self.name = name
        self.converter = converter


# Source for this list: https://www.w3.org/TR/SVG2/styling.html#PresentationAttributes
properties_list: List[PropertyDescription] = [
    PropertyDescription(
        "alignment-baseline",
        _EnumValueConverter(
            [
                "baseline",
                "text-bottom",
                "alphabetic",
                "ideographic",
                "middle",
                "central",
                "mathematical",
                "text-top",
            ]
        ),
        "baseline",
        True,
        False,
    ),
    PropertyDescription("baseline-shift", _StyleConverter(), "0", True, False),
    PropertyDescription("clip", _StyleConverter(), "auto", True, False),
    PropertyDescription("clip-path", _URLNoneValueConverter(), "none", True, False),
    PropertyDescription(
        "clip-rule", _EnumValueConverter(["nonzero", "evenodd"]), "nonzero", True, True
    ),
    PropertyDescription("color", _PaintValueConverter(), "black", True, True),
    PropertyDescription(
        "color-interpolation",
        _EnumValueConverter(["sRGB", "auto", "linearRGB"]),
        "sRGB",
        True,
        True,
    ),
    PropertyDescription(
        "color-interpolation-filters",
        _EnumValueConverter(["auto", "sRGB", "linearRGB"]),
        "linearRGB",
        True,
        True,
    ),
    PropertyDescription(
        "color-rendering",
        _EnumValueConverter(
            ["auto", "optimizeSpeed", "optimizeQuality", "pixelated", "crisp-edges"]
        ),
        "auto",
        True,
        True,
    ),
    PropertyDescription("cursor", _StyleConverter(), "auto", True, True),
    PropertyDescription(
        "direction", _EnumValueConverter(["ltr", "rtl"]), "ltr", True, True
    ),
    PropertyDescription(
        "display",
        _EnumValueConverter(
            [
                "inline",
                "block",
                "list-item",
                "inline-block",
                "table",
                "inline-table",
                "table-row-group",
                "table-header-group",
                "table-footer-group",
                "table-row",
                "table-column-group",
                "table-column",
                "table-cell",
                "table-caption",
                "none",
            ]
        ),
        "inline",
        True,
        False,
    ),
    PropertyDescription(
        "dominant-baseline",
        _EnumValueConverter(
            [
                "auto",
                "text-bottom",
                "alphabetic",
                "ideographic",
                "middle",
                "central",
                "mathematical",
                "hanging",
                "text-top",
            ]
        ),
        "auto",
        True,
        True,
    ),
    PropertyDescription("fill", _PaintValueConverter(), "black", True, True),
    PropertyDescription("fill-opacity", _AlphaValueConverter(), "1", True, True),
    PropertyDescription(
        "fill-rule", _EnumValueConverter(["nonzero", "evenodd"]), "nonzero", True, True
    ),
    PropertyDescription("filter", _FilterListConverter(), "none", True, False),
    PropertyDescription("flood-color", _PaintValueConverter(), "black", True, False),
    PropertyDescription("flood-opacity", _AlphaValueConverter(), "1", True, False),
    PropertyDescription("font-family", _StyleConverter(), "sans-serif", True, True),
    PropertyDescription("font-size", _FontSizeValueConverter(), "medium", True, True),
    PropertyDescription("font-size-adjust", _StyleConverter(), "none", True, True),
    PropertyDescription("font-stretch", _StyleConverter(), "normal", True, True),
    PropertyDescription(
        "font-style",
        _EnumValueConverter(["normal", "italic", "oblique"]),
        "normal",
        True,
        True,
    ),
    PropertyDescription(
        "font-variant",
        _EnumValueConverter(["normal", "small-caps"]),
        "normal",
        True,
        True,
    ),
    PropertyDescription(
        "font-weight",
        _EnumValueConverter(
            ["normal", "bold"] + [str(i) for i in range(100, 901, 100)]
        ),
        "normal",
        True,
        True,
    ),
    PropertyDescription(
        "glyph-orientation-horizontal", _StyleConverter(), "0deg", True, True
    ),
    PropertyDescription(
        "glyph-orientation-vertical", _StyleConverter(), "auto", True, True
    ),
    PropertyDescription("inline-size", _StyleConverter(), "0", False, False),
    PropertyDescription(
        "image-rendering",
        _EnumValueConverter(["auto", "optimizeQuality", "optimizeSpeed"]),
        "auto",
        True,
        True,
    ),
    PropertyDescription("letter-spacing", _StyleConverter(), "normal", True, True),
    PropertyDescription(
        "lighting-color", _ColorValueConverter(), "normal", True, False
    ),
    PropertyDescription("line-height", _StyleConverter(), "normal", False, True),
    PropertyDescription("marker", _MarkerShorthandValueConverter(), ""),
    PropertyDescription("marker-end", _URLNoneValueConverter(), "none", True, True),
    PropertyDescription("marker-mid", _URLNoneValueConverter(), "none", True, True),
    PropertyDescription("marker-start", _URLNoneValueConverter(), "none", True, True),
    PropertyDescription("mask", _URLNoneValueConverter(), "none", True, False),
    PropertyDescription("opacity", _AlphaValueConverter(), "1", True, False),
    PropertyDescription(
        "overflow",
        _EnumValueConverter(["visible", "hidden", "scroll", "auto"]),
        "visible",
        True,
        False,
    ),
    PropertyDescription("paint-order", _StyleConverter(), "normal", True, False),
    PropertyDescription(
        "pointer-events",
        _EnumValueConverter(
            [
                "bounding-box",
                "visiblePainted",
                "visibleFill",
                "visibleStroke",
                "visible",
                "painted",
                "fill",
                "stroke",
                "all",
                "none",
            ]
        ),
        "visiblePainted",
        True,
        True,
    ),
    PropertyDescription("shape-inside", _URLNoneValueConverter(), "none", False, False),
    PropertyDescription(
        "shape-rendering",
        _EnumValueConverter(
            ["auto", "optimizeSpeed", "crispEdges", "geometricPrecision"]
        ),
        "visiblePainted",
        True,
        True,
    ),
    PropertyDescription("stop-color", _ColorValueConverter(), "black", True, False),
    PropertyDescription("stop-opacity", _AlphaValueConverter(), "1", True, False),
    PropertyDescription("stroke", _PaintValueConverter(), "none", True, True),
    PropertyDescription(
        "stroke-dasharray", _StrokeDasharrayValueConverter(), "none", True, True
    ),
    PropertyDescription("stroke-dashoffset", _StyleConverter(), "0", True, True),
    PropertyDescription(
        "stroke-linecap",
        _EnumValueConverter(["butt", "round", "square"]),
        "butt",
        True,
        True,
    ),
    PropertyDescription(
        "stroke-linejoin",
        _EnumValueConverter(["miter", "miter-clip", "round", "bevel", "arcs"]),
        "miter",
        True,
        True,
    ),
    PropertyDescription("stroke-miterlimit", _StyleConverter(), "4", True, True),
    PropertyDescription("stroke-opacity", _AlphaValueConverter(), "1", True, True),
    PropertyDescription("stroke-width", _StyleConverter(), "1", True, True),
    PropertyDescription("text-align", _StyleConverter(), "start", True, True),
    PropertyDescription(
        "text-anchor",
        _EnumValueConverter(["start", "middle", "end"]),
        "start",
        True,
        True,
    ),
    PropertyDescription(
        "text-decoration-line", _StyleConverter(), "none", False, False
    ),
    PropertyDescription(
        "text-decoration-style",
        _EnumValueConverter(["solid", "double", "dotted", "dashed", "wavy"]),
        "solid",
        False,
        False,
    ),
    PropertyDescription(
        "text-decoration-color", _StyleConverter(), "currentcolor", False, False
    ),
    PropertyDescription(
        "text-overflow", _EnumValueConverter(["clip", "ellipsis"]), "clip", True, False
    ),
    PropertyDescription(
        "text-rendering",
        _EnumValueConverter(
            ["auto", "optimizeSpeed", "optimizeLegibility", "geometricPrecision"]
        ),
        "auto",
        True,
        True,
    ),
    PropertyDescription(
        "unicode-bidi",
        _EnumValueConverter(
            [
                "normal",
                "embed",
                "isolate",
                "bidi-override",
                "isolate-override",
                "plaintext",
            ]
        ),
        "normal",
        True,
        False,
    ),
    PropertyDescription("vector-effect", _StyleConverter(), "none", True, False),
    PropertyDescription("vertical-align", _StyleConverter(), "baseline", False, False),
    PropertyDescription(
        "visibility",
        _EnumValueConverter(["visible", "hidden", "collapse"]),
        "visible",
        True,
        True,
    ),
    PropertyDescription(
        "white-space",
        _EnumValueConverter(
            ["normal", "pre", "nowrap", "pre-wrap", "break-spaces", "pre-line"]
        ),
        "normal",
        True,
        True,
    ),
    PropertyDescription("word-spacing", _StyleConverter(), "normal", True, True),
    PropertyDescription(
        "writing-mode",
        _EnumValueConverter(
            [
                "horizontal-tb",
                "vertical-rl",
                "vertical-lr",
                "lr",
                "lr-tb",
                "rl",
                "rl-tb",
                "tb",
                "tb-rl",
            ]
        ),
        "horizontal-tb",
        True,
        True,
    ),
    PropertyDescription(
        "-inkscape-font-specification", _StyleConverter(), "sans-serif", False, True
    ),
]
all_properties = {v.name: v for v in properties_list}

properties_list += [
    PropertyDescription("font", _FontValueShorthandConverter(), ""),
    PropertyDescription("text-decoration", _TextDecorationValueConverter(), ""),
]

all_properties["font"] = properties_list[-2]
all_properties["text-decoration"] = properties_list[-1]

shorthand_from_value = {
    item: prop.name
    for prop in properties_list
    if isinstance(prop.converter, _ShorthandValueConverter)
    for item in prop.converter.keys
}

shorthand_properties = {
    i.name: i
    for i in properties_list
    if isinstance(i.converter, _ShorthandValueConverter)
}



"""
inkvn Reader

Reads Linearity Curve / Vectornator files and convert them into intermediate data.
"""

import logging
import zipfile
from typing import List

from packaging import version

import inkvn.reader.decode as d
import inkvn.reader.decode_vn as dvn
import inkvn.reader.extract as ext
from ..elements.artboard import VNArtboard


class CurveReader:
    """
    inkvn CurveReader

    A Linearity Curve / Vectornator file reader to convert Curve documents into dataclasses.
    """

    def __init__(self, stream):
        self.archive = zipfile.ZipFile(stream, "r")
        self.file_version: int = 44  # main support
        self.app_version: str
        self.artboards: List[VNArtboard] = []

        self.read()

    def read(self):
        manifest = ext.extract_manifest(self.archive)
        document = ext.extract_document(self.archive, manifest)
        drawing_data = ext.extract_drawing_data(document)

        units = drawing_data["settings"]["units"]
        self.app_version = document["appVersion"]
        self.file_version = manifest["fileFormatVersion"]
        artboard_paths = drawing_data["artboardPaths"]

        # different file versions have incompatible structure, reporting App version & File version greatly helps
        # inkex.utils.debug(f"App version: {self.app_version}, File version: {self.file_version}, File name: {self.archive.filename}")

        assert len(artboard_paths), "No artboard paths found in the document."

        # will be used later (as Inkscape attribute)
        # print(f"Unit: {units}")

        # Read Artboard (GUID JSON)
        for artboard_path in artboard_paths:
            gid_json = ext.extract_gid_json(self.archive, artboard_path)

            # if the file is Linearity Curve
            if self.check_if_curve(self.app_version):
                artboard = d.read_artboard(self.archive, gid_json)
                self.artboards.append(artboard)

            # if the file is Vectornator
            else:
                artboard = dvn.read_vn_artboard(self.archive, gid_json)
                self.artboards.append(artboard)

    @staticmethod
    def check_if_curve(input_version: str):
        """check if the app version is 5.x or not"""
        required_version = version.parse("5.1.0")
        try:
            current_version = version.parse(input_version)
        except version.InvalidVersion:
            logging.warning(f"Invalid version string: {input_version}")
            return False

        return current_version >= required_version
